#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
  AbstractJob
  ^        ^
  |        |
  +        +
Job       AbstractSpecificJob
            ^             ^
            |             |
            +             +
        ModelJob       PredictJob


Subclasses of AbstractJob can override _converter_extra and _extra_fields to indicate extra object
attributes which come in the job data. They must implement the methods which raise
NotImplementedError

The Job class represents jobs of any type (whether or not we have classes for more specific job
types) with only the data and methods that work for jobs of any type.

We also have classes representing specific jobs. Some of the functionality for these jobs is shared
and hence is implemented in AbstractSpecificJob.

AbstractSpecificJob and AbstractJob should be non-public and may change. Users should only rely
on the concrete classes.
"""
from __future__ import annotations

from typing import Any, Dict, Optional, Tuple, Type, TypeVar

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.shap_impact import ShapImpact
from datarobot.models.validators import feature_impact_trafaret, single_feature_impact_trafaret
from datarobot.utils.waiters import wait_for_async_resolution

from .. import errors
from ..client import get_client, staticproperty
from ..enums import DEFAULT_MAX_WAIT, JOB_TYPE, PREDICTION_PREFIX, QUEUE_STATUS
from ..utils import from_api, raw_prediction_response_to_dataframe
from .feature_effect import FeatureEffects, FeatureEffectsMulticlass
from .model import DatetimeModel, Model, PrimeModel
from .prediction_explanations import PredictionExplanations, PredictionExplanationsInitialization
from .prime_file import PrimeFile
from .rating_table import RatingTable
from .ruleset import Ruleset
from .training_predictions import TrainingPredictions

TAbstractSpecificJob = TypeVar("TAbstractSpecificJob", bound="AbstractSpecificJob")


class AbstractJob:
    """Generic representation of a job in the queue"""

    # Subclasses can override this:
    _converter_extra = t.Dict()
    _extra_fields = frozenset()

    # Subclasses should not override any of this:
    _client = staticproperty(get_client)
    _converter_common = t.Dict(
        {
            t.Key("id", optional=True): Int,
            t.Key("status", optional=True): t.Enum(
                QUEUE_STATUS.ABORTED,
                QUEUE_STATUS.COMPLETED,
                QUEUE_STATUS.ERROR,
                QUEUE_STATUS.INPROGRESS,
                QUEUE_STATUS.QUEUE,
            ),
            t.Key("project_id", optional=True): String,
            t.Key("model_id", optional=True): String,
            t.Key("is_blocked"): t.Bool,
        }
    )

    def __init__(self, data: Dict[str, Any], completed_resource_url: Optional[str] = None) -> None:
        # Importing here to dodge circular dependency
        from . import Project  # pylint: disable=import-outside-toplevel,cyclic-import

        if not isinstance(data, dict):
            raise ValueError("job data must be a dict")
        self._completed_resource_url = completed_resource_url
        data = from_api(data)

        converter = (self._converter_common + self._converter_extra).allow_extra("*")
        self._safe_data = converter.check(data)
        self.job_type = self._get_job_type(self._safe_data)
        self.status = self._safe_data.get("status")
        self.id = self._safe_data.get("id")
        self.project = Project(self._safe_data.get("project_id"))
        self.project_id = self._safe_data.get("project_id")
        self.model_id = self._safe_data.get("model_id")
        self.model = Model(self.model_id) if self.model_id else None
        self.is_blocked = self._safe_data.get("is_blocked")
        for k in self._extra_fields:
            v = self._safe_data.get(k)
            setattr(self, k, v)

    @classmethod
    def _get_job_type(cls, safe_data):
        raise NotImplementedError

    def __repr__(self):
        return f"Job({self.job_type}, status={self.status})"

    @classmethod
    def _data_and_completed_url_for_job(  # pylint: disable=missing-function-docstring
        cls,
        url: str,
    ) -> Tuple[Dict[str, Any], str]:
        response = cls._client.get(url, allow_redirects=False)

        if response.status_code in (200, 303):
            data = response.json()
            completed_url = response.headers["Location"] if response.status_code == 303 else None
            return data, completed_url
        else:
            e_msg = "Server unexpectedly returned status code {}"
            raise errors.AsyncFailureError(e_msg.format(response.status_code))

    def refresh(self):
        """
        Update this object with the latest job data from the server.
        """
        data, completed_url = self._data_and_completed_url_for_job(self._this_job_path())
        self.__init__(  # pylint: disable=unnecessary-dunder-call
            data, completed_resource_url=completed_url
        )

    def get_result(self, params=None):
        """
        Parameters
        ----------
        params : dict or None
            Query parameters to be added to request to get results.

        Notes
        -----
        For featureEffects, source param is required to define source,
        otherwise the default is `training`.

        Returns
        -------
        result : object
            Return type depends on the job type
                - for model jobs, a `Model` is returned
                - for predict jobs, a `pandas.DataFrame` (with predictions) is returned
                - for featureImpact jobs, a list of dicts by default (see ``with_metadata``
                  parameter of the ``FeatureImpactJob`` class and its ``get()`` method).
                - for primeRulesets jobs, a list of `Rulesets`
                - for primeModel jobs, a `PrimeModel`
                - for primeDownloadValidation jobs, a `PrimeFile`
                - for predictionExplanationInitialization jobs, a `PredictionExplanationsInitialization`
                - for predictionExplanations jobs, a `PredictionExplanations`
                - for featureEffects, a `FeatureEffects`.

        Raises
        ------
        JobNotFinished
            If the job is not finished, the result is not available.
        AsyncProcessUnsuccessfulError
            If the job errored or was aborted

        """
        self.refresh()
        if self.status in [QUEUE_STATUS.ERROR, QUEUE_STATUS.ABORTED]:
            raise errors.AsyncProcessUnsuccessfulError
        if not self._completed_resource_url:
            raise errors.JobNotFinished
        completed_resource_path = self._client.strip_endpoint(self._completed_resource_url)
        return self._make_result_from_location(completed_resource_path, params)

    def _make_result_from_json(self, server_data):  # pylint: disable=missing-function-docstring
        if self.job_type == JOB_TYPE.MODEL:
            if "/datetimeModels" in self._completed_resource_url:
                return DatetimeModel.from_server_data(server_data)
            return Model.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.PREDICT:
            return raw_prediction_response_to_dataframe(server_data, PREDICTION_PREFIX.DEFAULT)
        elif self.job_type == JOB_TYPE.FEATURE_IMPACT:
            # Note: a custom FeatureImpactJob class is used for high level API now.
            return server_data["featureImpacts"]
        elif self.job_type == JOB_TYPE.FEATURE_EFFECTS:
            use_insights_format = "count" in server_data
            return FeatureEffects.from_server_data(
                server_data, use_insights_format=use_insights_format
            )
        elif self.job_type == JOB_TYPE.PRIME_RULESETS:
            return [Ruleset.from_server_data(ruleset_data) for ruleset_data in server_data]
        elif self.job_type == JOB_TYPE.PRIME_MODEL:
            return PrimeModel.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.PRIME_VALIDATION:
            return PrimeFile.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.PREDICTION_EXPLANATIONS_INITIALIZATION:
            return PredictionExplanationsInitialization.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.RATING_TABLE_VALIDATION:
            return RatingTable.from_server_data(server_data)
        elif self.job_type == JOB_TYPE.SHAP_IMPACT:
            return ShapImpact.from_server_data(server_data)
        else:
            raise ValueError(f"Unrecognized job type {self.job_type}.")

    def _make_result_from_location(
        self, location, params=None
    ):  # pylint: disable=missing-function-docstring
        if self.job_type == JOB_TYPE.TRAINING_PREDICTIONS:
            return TrainingPredictions.from_location(location)

        if self.job_type == JOB_TYPE.PREDICTION_EXPLANATIONS:
            head, tail = location.split("/predictionExplanations/", 1)
            project_id, prediction_expl_id = head.split("/")[-1], tail.split("/")[0]
            return PredictionExplanations.get(
                project_id=project_id, prediction_explanations_id=prediction_expl_id
            )

        server_data = self._client.get(location, params=params).json()
        return self._make_result_from_json(server_data)

    def wait_for_completion(self, max_wait: int = DEFAULT_MAX_WAIT) -> None:
        """
        Waits for job to complete.

        Parameters
        ----------
        max_wait : Optional[int]
            How long to wait for the job to finish.
        """
        try:
            wait_for_async_resolution(self._client, self._this_job_path(), max_wait=max_wait)
        finally:
            # We are gonna try to update the job data, that's OK if it fails too (rare cases)
            self.refresh()

    def get_result_when_complete(self, max_wait=DEFAULT_MAX_WAIT, params=None):
        """
        Parameters
        ----------
        max_wait : Optional[int]
            How long to wait for the job to finish.

        params : dict, optional
            Query parameters to be added to request.

        Returns
        -------
        result: object
            Return type is the same as would be returned by `Job.get_result`.

        Raises
        ------
        AsyncTimeoutError
            If the job does not finish in time
        AsyncProcessUnsuccessfulError
            If the job errored or was aborted
        """
        if self.job_type == JOB_TYPE.MODEL_EXPORT:
            # checking this here instead of in _make_result to avoid waiting for the response
            raise ValueError(
                "Can't return the result for a model export job. Use "
                "Job.wait_for_completion to wait for the job to complete and "
                "Model.download_export to download the finished export."
            )

        self.wait_for_completion(max_wait=max_wait)
        return self._make_result_from_location(self._completed_resource_url, params=params)

    def _this_job_path(self) -> str:
        return self._job_path(self.project.id, self.id)

    @classmethod
    def _job_path(cls, project_id, job_id):
        raise NotImplementedError

    @classmethod
    def get(cls, project_id, job_id):
        # Note: For 3.0 (when the behavior of all job types' `get` methods can be made consistent),
        #       the implementation can move here.
        raise NotImplementedError

    def cancel(self):
        """
        Cancel this job. If this job has not finished running, it will be
        removed and canceled.
        """
        self._client.delete(self._this_job_path())


class Job(AbstractJob):
    """Tracks asynchronous work being done within a project

    Attributes
    ----------
    id : int
        the id of the job
    project_id : str
        the id of the project the job belongs to
    status : str
        the status of the job - will be one of ``datarobot.enums.QUEUE_STATUS``
    job_type : str
        what kind of work the job is doing - will be one of ``datarobot.enums.JOB_TYPE``
    is_blocked : bool
        if true, the job is blocked (cannot be executed) until its dependencies are resolved
    """

    _converter_extra = t.Dict(
        {t.Key("job_type", optional=True) >> "job_type": String, t.Key("url") >> "url": String}
    )

    def __init__(self, data: Dict[str, Any], completed_resource_url: Optional[str] = None) -> None:
        super().__init__(data, completed_resource_url=completed_resource_url)
        self._job_details_path = self._client.strip_endpoint(self._safe_data["url"])

    @classmethod
    def _get_job_type(cls, safe_data):
        # For generic jobs, job_type comes from the data.
        return safe_data.get("job_type")

    @classmethod
    def _job_path(cls, project_id: str, job_id: str) -> str:
        # Path where you can get jobs of this kind.
        return f"projects/{project_id}/jobs/{job_id}/"

    @classmethod
    def get(cls, project_id: str, job_id: str) -> Job:
        """
        Fetches one job.

        Parameters
        ----------
        project_id : str
            The identifier of the project in which the job resides
        job_id : str
            The job id

        Returns
        -------
        job : Job
            The job

        Raises
        ------
        AsyncFailureError
            Querying this resource gave a status code other than 200 or 303
        """
        url = cls._job_path(project_id, job_id)
        data, completed_url = cls._data_and_completed_url_for_job(url)
        return cls(data, completed_resource_url=completed_url)


class AbstractSpecificJob(AbstractJob):  # pylint: disable=missing-class-docstring
    @classmethod
    def _job_type(cls):
        raise NotImplementedError

    @classmethod
    def _get_job_type(cls, _):
        return cls._job_type()

    @classmethod
    def from_job(cls, job):  # pylint: disable=missing-function-docstring
        if not job.job_type == cls._job_type():
            raise ValueError(f"wrong job_type: {job.job_type}")
        if isinstance(job, AbstractSpecificJob):
            job.refresh()
            return job
        else:
            response = cls._client.get(job._job_details_path)
            data = response.json()
            return cls(data)

    @classmethod
    def from_id(
        cls: Type[TAbstractSpecificJob], project_id: str, job_id: str
    ) -> TAbstractSpecificJob:
        url = cls._job_path(project_id, job_id)
        response = cls._client.get(url, allow_redirects=False)
        data = response.json()
        return cls(data)

    @classmethod
    def get(cls: Type[TAbstractSpecificJob], project_id: str, job_id: str) -> TAbstractSpecificJob:
        # Note: In v3.0 the desired behavior here will be the same as in Job, so we can delete this
        url = cls._job_path(project_id, job_id)
        response = cls._client.get(url, allow_redirects=False)
        if response.status_code == 200:
            data = response.json()
            return cls(data)
        elif response.status_code == 303:
            raise errors.PendingJobFinished
        else:
            e_msg = "Server unexpectedly returned status code {}"
            raise errors.AsyncFailureError(e_msg.format(response.status_code))


class TrainingPredictionsJob(Job):  # pylint: disable=missing-class-docstring
    def __init__(self, data, model_id, data_subset, **kwargs):
        super().__init__(data, **kwargs)
        self._model_id = model_id
        self._data_subset = data_subset

    @classmethod
    def get(cls, project_id, job_id, model_id=None, data_subset=None):
        """
        Fetches one training predictions job.

        The resulting
        :py:class:`TrainingPredictions <datarobot.models.training_predictions.TrainingPredictions>`
        object will be annotated with `model_id` and `data_subset`.

        Parameters
        ----------
        project_id : str
            The identifier of the project in which the job resides
        job_id : str
            The job id
        model_id : str
            The identifier of the model used for computing training predictions
        data_subset : dr.enums.DATA_SUBSET, optional
            Data subset used for computing training predictions

        Returns
        -------
        job : TrainingPredictionsJob
            The job
        """
        url = cls._job_path(project_id, job_id)
        data, completed_url = cls._data_and_completed_url_for_job(url)
        return cls(
            data,
            model_id=model_id,
            data_subset=data_subset,
            completed_resource_url=completed_url,
        )

    def _make_result_from_location(self, location, params=None):
        return TrainingPredictions.from_location(
            location,
            model_id=self._model_id,
            data_subset=self._data_subset,
        )

    def refresh(self):
        """
        Update this object with the latest job data from the server.
        """
        data, completed_url = self._data_and_completed_url_for_job(self._this_job_path())
        self.__init__(  # pylint: disable=unnecessary-dunder-call
            data,
            model_id=self._model_id,
            data_subset=self._data_subset,
            completed_resource_url=completed_url,
        )


class FeatureImpactJob(Job):
    """Custom Feature Impact job to handle different return value structures.

    The original implementation had just the the data and the new one also includes some metadata.

    In general, we aim to keep the number of Job classes low by just utilizing the `job_type`
    attribute to control any specific formatting; however in this case when we needed to support
    a new representation with the _same_ job_type, customizing the behavior of
    `_make_result_from_location` allowed us to achieve our ends without complicating the
    `_make_result_from_json` method.
    """

    # This is a default value used by a new instantiated job. This can be overridden in `__init__()`
    # but for existing instances it will not change on subsequent `__init__()` calls without
    # explicit `with_metadata` parameter (which `refresh()` does).
    _with_metadata = False

    def __init__(self, data, completed_resource_url=None, with_metadata=False):
        super().__init__(data, completed_resource_url=completed_resource_url)
        # We might be in existing instance with .refresh() called, that does not know about
        # with_metadata parameter. Only set instance attribute if it is set to True explicitly,
        # otherwise default to class attribute that is False.
        if with_metadata:
            self._with_metadata = with_metadata

    @classmethod
    def get(cls, project_id, job_id, with_metadata=False):
        """
        Fetches one job.

        Parameters
        ----------
        project_id : str
            The identifier of the project in which the job resides
        job_id : str
            The job id
        with_metadata : bool
            To make this job return the metadata (i.e. the full object of the completed resource)
            set the `with_metadata` flag to True.

        Returns
        -------
        job : Job
            The job

        Raises
        ------
        AsyncFailureError
            Querying this resource gave a status code other than 200 or 303
        """
        url = cls._job_path(project_id, job_id)
        data, completed_url = cls._data_and_completed_url_for_job(url)
        return cls(data, completed_resource_url=completed_url, with_metadata=with_metadata)

    def _make_result_from_location(self, location, params=None):
        """Custom extension of get_result_when_complete to use a customized formatter"""
        server_data = self._client.get(location, params=params).json()
        return filter_feature_impact_result(server_data, with_metadata=self._with_metadata)


class FeatureEffectsMulticlassJob(Job):
    def _make_result_from_location(self, location, params=None):
        return FeatureEffectsMulticlass.from_location(location)


def _filter_feature_impact_result(data, validator_config):
    # The trafaret used might allow extra fields, but we want this method to return only known
    # fields. Be conservative in what you do, be liberal in what you accept from others.
    exposed_fields = [k.name for k in validator_config.keys]
    return {k: v for k, v in data.items() if k in exposed_fields}


def filter_feature_impact_result(data, with_metadata):
    """Filter Feature Impact response according to specific validator configuration.

    Parameters
    ----------
    data : dict
        A (previously validated) data to filter.
    with_metadata : bool
        The flag indicating if the result should include the metadata as well.

    Returns
    -------
    dict or list
        A filtered data.
    """
    if not with_metadata:
        return [
            _filter_feature_impact_result(item, single_feature_impact_trafaret)
            for item in data["featureImpacts"]
        ]
    return _filter_feature_impact_result(data, feature_impact_trafaret)



#
# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from ..errors import UpdateAttributesError


class UpdateAttributesMixin:
    """A mixin to allow updating all attributes on an instance"""

    def _update_attributes(self, **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Update individual attributes of an instance of this class.

        Raises
        ------
        UpdateAttributesError
            Raised if a kwarg was passed that doesn't correspond to an
            attribute on this class.
        """
        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                raise UpdateAttributesError(
                    class_name=type(self).__name__,
                    invalid_key=key,
                    message=f"This {type(self).__name__} instance does not contain attribute `{key}`.",
                )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Dict, Iterable, List, Mapping, Optional

from mypy_extensions import TypedDict
import trafaret as t

from datarobot._compat import String
from datarobot.models.api_object import APIObject

LiftChartBinsTrafaret = t.Dict(
    {
        t.Key("bins"): t.List(
            t.Dict(
                {
                    t.Key("actual"): t.Float,
                    t.Key("predicted"): t.Float,
                    t.Key("bin_weight"): t.Float,
                }
            ).ignore_extra("*")
        )
    }
)


class LiftChartBin(TypedDict):
    actual: float
    predicted: float
    bin_weight: float


class LiftChartData(TypedDict):
    bins: List[LiftChartBin]


class LiftChartServerData(TypedDict):
    id: str
    entity_id: str
    project_id: str
    source: str
    data_slice_id: Optional[str]
    data: LiftChartData


class LiftChart(APIObject):
    """ Lift chart data for model.

    Notes
    -----
    ``LiftChartBin`` is a dict containing the following:

        * ``actual`` (float) Sum of actual target values in bin
        * ``predicted`` (float) Sum of predicted target values in bin
        * ``bin_weight`` (float) The weight of the bin. For weighted projects, it is the sum of \
          the weights of the rows in the bin. For unweighted projects, it is the number of rows in \
          the bin.

    Attributes
    ----------
    source : str
        Lift chart data source. Can be 'validation', 'crossValidation' or 'holdout'.
    bins : list of dict
        List of dicts with schema described as ``LiftChartBin`` above.
    source_model_id : str
        ID of the model this lift chart represents; in some cases,
        insights from the parent of a frozen model may be used
    target_class : Optional[str]
        For multiclass lift - target class for this lift chart data.
    data_slice_id: string or None
        The slice to retrieve Lift Chart for; if None, retrieve unsliced data.
    """

    _converter = (
        t.Dict(
            {
                t.Key("source"): String,
                t.Key("source_model_id"): String,
                t.Key("target_class", optional=True, default=None): t.Or(String, t.Null),
                t.Key("data_slice_id", optional=True): t.Or(String, t.Null),
            }
        )
        .merge(LiftChartBinsTrafaret)
        .ignore_extra("*")
    )

    def __init__(self, source, bins, source_model_id, target_class, data_slice_id=None):
        self.source = source
        self.bins = bins
        self.source_model_id = source_model_id
        self.target_class = target_class
        self.data_slice_id = data_slice_id

    def __repr__(self):
        additional_params = ""
        if self.target_class:
            additional_params += f"{self.target_class}:"
        if self.data_slice_id:
            additional_params += f"{self.data_slice_id}:"
        return f"LiftChart({additional_params}{self.source})"

    @staticmethod
    def _repack_insights_response(server_data: LiftChartServerData):
        """Repack the JSON sent by the GET /insights/ endpoint to match the format expected by the
        insight APIObject class.

        Parameters
        ----------
        server_data : dict
        {
            "id": "6474726c7ca961a4ebce068d",
            "entityId": "64747242956c7390bb15b206",
            "projectId": "647471b6b5d9cbd454f2cf63",
            "source": "validation",
            "dataSliceId": "647471b6b5d9cbd454f2ab99",
            "data": {"bins": [{"actual":0.0, "predicted":0.0, "binWeight":27.0}, ...]}
        }

        Returns
        -------
        dict
        {
            "source": "validation",
            "bins": [{"actual":0.0, "predicted":0.0, "binWeight":27.0}, ...],
            "sourceModelId": "64747242956c7390bb15b206",
            "dataSliceId": "647471b6b5d9cbd454f2ab99",
        }
        """
        return {
            "source": server_data["source"],
            "bins": server_data["data"]["bins"],
            "sourceModelId": server_data["entityId"],
            "dataSliceId": server_data["dataSliceId"],
        }

    @classmethod
    def from_server_data(cls, data, keep_attrs=None, use_insights_format=False, **kwargs):
        """
        Overwrite APIObject.from_server_data to handle lift chart data retrieved
        from either legacy URL or /insights/ new URL.

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        use_insights_format : Optional[bool]
            Whether to repack the data from the format used in the GET /insights/liftChart/ URL
            to the format used in the legacy URL.
        """
        if use_insights_format:
            data = cls._repack_insights_response(data)

        return super().from_server_data(data=data, keep_attrs=keep_attrs)


class SlicedLiftChart(LiftChart):
    """Wrapper around LiftChart to override `from_server_data` method"""

    @classmethod
    # type: ignore[override]
    def from_server_data(
        cls,
        data: Dict[str, Any],
        keep_attrs: Optional[Iterable[str]] = None,
        use_insights_format: bool = True,
        **kwargs: Mapping[str, Any],
    ) -> "LiftChart":
        """
        Overwrite LiftChart.from_server_data to set `use_insights_format=True` by default
        This is necessary for the correct transformation of data received from /insights endpoints

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place.
        keep_attrs : iterable
            List, set or tuple of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        use_insights_format : Optional[bool]
            Whether to repack the data from the format used in the GET /insights/liftChart/ URL
            to the format used in the legacy URL.
        """
        if data.get("count"):
            # it's a list
            data = data["data"][0]

        return super().from_server_data(
            data=data, keep_attrs=keep_attrs, use_insights_format=use_insights_format, **kwargs
        )



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Optional

from mypy_extensions import TypedDict


class UserMetadata(TypedDict):
    id: str
    email: Optional[str]
    name: Optional[str]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, List

import trafaret as t

from datarobot._compat import String
from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.models.api_object import APIObject
from datarobot.utils import pagination
from datarobot.utils.waiters import wait_for_async_resolution

# Numeric features.
numeric_insight = t.Dict(
    {t.Key("statistic", optional=True): t.Float(), t.Key("cluster_name"): String()}
)

numeric_insight_per_cluster = t.Dict(
    {
        t.Key("all_data", optional=True): t.Float(),
        t.Key("per_cluster"): t.List(numeric_insight),
        t.Key("insight_name"): String(),
    }
)

# Categorical features.

categorical_per_value_statistic = t.Dict(
    {t.Key("category_level"): String(), t.Key("frequency"): t.Float()}
)

categorical_per_value_statistic_list = t.Dict(
    {
        t.Key("all_other"): t.Float(),
        t.Key("missing_rows_percent"): t.Float(),
        t.Key("per_value_statistics"): t.List(categorical_per_value_statistic),
        t.Key("cluster_name"): String(),
    }
)

categorical_insight_per_cluster = t.Dict(
    {
        t.Key("all_data"): t.Dict(
            {
                t.Key("all_other"): t.Float(),
                t.Key("missing_rows_percent"): t.Float(),
                t.Key("per_value_statistics"): t.List(categorical_per_value_statistic),
            }
        ),
        t.Key("per_cluster"): t.List(categorical_per_value_statistic_list),
        t.Key("insight_name"): String(),
    }
)

# Text feature.

text_per_ngram_statistic = t.Dict(
    {
        t.Key("ngram"): String(),
        t.Key("importance"): t.Float(),
        t.Key("contextual_extracts"): t.List(String()),
    }
)

text_per_ngram_statistic_list = t.Dict(
    {
        t.Key("missing_rows_percent"): t.Or(t.Float(), t.Null),
        t.Key("per_value_statistics"): t.List(text_per_ngram_statistic),
        t.Key("cluster_name"): String(),
    }
)

text_important_ngram_insight_per_cluster = t.Dict(
    {
        t.Key("all_data"): t.Dict(
            {
                t.Key("missing_rows_percent"): t.Or(t.Float(), t.Null),
                t.Key("per_value_statistics"): t.List(text_per_ngram_statistic),
            }
        ),
        t.Key("per_cluster"): t.List(text_per_ngram_statistic_list),
        t.Key("insight_name"): String(),
    }
)

# Image feature.

image_insight_statistic = t.Dict(
    {
        t.Key("images"): t.List(String()),
        t.Key("percentage_of_missing_images"): t.Float(),
        t.Key("cluster_name"): String(),
    }
)

image_insight_per_cluster = t.Dict(
    {
        t.Key("all_data"): t.Dict(
            {
                t.Key("image_entities"): t.List(String()),
                t.Key("percentage_of_missing_images"): t.Float(),
            }
        ),
        t.Key("per_cluster"): t.List(image_insight_statistic),
        t.Key("insight_name"): String(),
    }
)

# Geospatial feature.

geospatial_insight_values = t.Dict(
    {t.Key("representative_locations"): t.List(t.List(t.Float())), t.Key("cluster_name"): String()}
)

geospatial_insight_per_cluster = t.Dict(
    {t.Key("per_cluster"): t.List(geospatial_insight_values), t.Key("insight_name"): String()}
)


class ClusterInsight(APIObject):
    """Holds data on all insights related to feature as well as breakdown per cluster.

    Parameters
    ----------
    feature_name: str
        Name of a feature from the dataset.
    feature_type: str
        Type of feature.
    insights : List[ClusterInsight]
        List provides information regarding the importance of a specific feature in relation
        to each cluster. Results help understand how the model is grouping data and what each
        cluster represents.
    feature_impact: float
        Impact of a feature ranging from 0 to 1.
    """

    _compute_path = "projects/{project_id}/models/{model_id}/clusterInsights/"
    _retrieve_path = "projects/{project_id}/models/{model_id}/clusterInsights/"
    _converter = t.Dict(
        {
            t.Key("feature_name"): String(),
            t.Key("feature_type"): String(),
            t.Key("feature_impact", optional=True): t.Float(),
            t.Key("insights"): t.List(
                t.Or(
                    numeric_insight_per_cluster,
                    categorical_insight_per_cluster,
                    text_important_ngram_insight_per_cluster,
                    image_insight_per_cluster,
                    geospatial_insight_per_cluster,
                )
            ),
        }
    ).ignore_extra("*")

    def __init__(self, **kwargs: Any) -> None:
        self.feature_name = kwargs.get("feature_name")
        self.feature_type = kwargs.get("feature_type")
        self.insights = kwargs.get("insights")
        self.feature_impact = kwargs.get("feature_impact")

    def __repr__(self) -> str:
        return (
            "ClusterInsight("
            "feature_name={0.feature_name}, "
            "feature_type={0.feature_type}, "
            "feature_impact={0.feature_impact})"
        ).format(self)

    @classmethod
    def list(cls, project_id: str, model_id: str) -> List[ClusterInsight]:
        path = cls._retrieve_path.format(project_id=project_id, model_id=model_id)
        return [cls.from_server_data(x) for x in pagination.unpaginate(path, {}, cls._client)]

    @classmethod
    def compute(
        cls, project_id: str, model_id: str, max_wait: int = DEFAULT_MAX_WAIT
    ) -> List[ClusterInsight]:
        """Starts creation of cluster insights for the model and if successful, returns computed
        ClusterInsights. This method allows calculation to continue for a specified time and
        if not complete, cancels the request.

        Parameters
        ----------
        project_id: str
            ID of the project to begin creation of cluster insights for.
        model_id: str
                        ID of the project model to begin creation of cluster insights for.
        max_wait: int
            Maximum number of seconds to wait canceling the request.

        Returns
        -------
        List[ClusterInsight]

        Raises
        ------
        ClientError
            Server rejected creation due to client error.
            Most likely cause is bad ``project_id`` or ``model_id``.
        AsyncFailureError
            Indicates whether any of the responses from the server are unexpected.
        AsyncProcessUnsuccessfulError
            Indicates whether the cluster insights computation failed or was cancelled.
        AsyncTimeoutError
            Indicates whether the cluster insights computation did not resolve within the specified
            time limit (max_wait).
        """
        compute_path = cls._compute_path.format(project_id=project_id, model_id=model_id)
        response = cls._client.post(compute_path)
        wait_for_async_resolution(cls._client, response.headers["Location"], max_wait=max_wait)
        return cls.list(project_id, model_id)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import abc
from datetime import datetime
import json

import pandas as pd
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import TARGET_TYPE
from datarobot.models.project_options import ProjectOptions

from ..utils import get_id_from_response
from .api_object import APIObject

int_float_string = t.Type(int) | t.Type(float) | String(allow_blank=True)

prediction_values_trafaret = t.Dict(
    {t.Key("label", optional=True): int_float_string, t.Key("value"): t.Float}
).ignore_extra("*")

per_ngram_text_explanations_trafaret = t.Or(
    t.List(
        t.Dict(
            {
                t.Key("ngrams"): t.List(
                    t.Dict({t.Key("ending_index"): Int, t.Key("starting_index"): Int})
                ),
                t.Key("is_unknown"): t.Bool,
                t.Key("qualitative_strength"): String,
                t.Key("strength"): t.Float,
            }
        )
    ),
    t.Null,
)

prediction_explanations_entry_trafaret = t.Dict(
    {
        t.Key("label", optional=True): int_float_string,
        t.Key("feature"): String,
        t.Key("feature_value"): int_float_string,
        t.Key("strength"): t.Float,
        t.Key("qualitative_strength"): String,
        t.Key("per_ngram_text_explanations", optional=True): per_ngram_text_explanations_trafaret,
    }
).ignore_extra("*")

prediction_explanations_trafaret = t.Dict(
    {
        t.Key("row_id"): Int,
        t.Key("prediction"): int_float_string,
        t.Key("adjusted_prediction", optional=True): int_float_string,
        t.Key("prediction_values"): t.List(prediction_values_trafaret),
        t.Key("adjusted_prediction_values", optional=True): t.List(prediction_values_trafaret),
        t.Key("prediction_explanations"): t.List(prediction_explanations_entry_trafaret),
    }
).ignore_extra("*")

KEEP_ATTRS = ["data.prediction_explanations.per_ngram_text_explanations"]


class PredictionExplanationsMode:
    """Prediction explanations mode for multiclass models"""

    @abc.abstractmethod
    def get_api_parameters(self, batch_route=False):
        """Get parameters passed in corresponding API call

        Parameters
        ----------
        batch_route : bool
            Batch routes describe prediction calls with all possible parameters, so to
            distinguish explanation parameters from others they have prefix in parameters.

        Returns
        -------
        dict
        """


class TopPredictionsMode(PredictionExplanationsMode):
    """Calculate prediction explanations for the number of top predicted classes in each row.

    Attributes
    ----------
    num_top_classes : int
        Number of top predicted classes [1..10] that will be explained for each dataset row.
    """

    def __init__(self, num_top_classes):
        self.num_top_classes = num_top_classes

    def get_api_parameters(self, batch_route=False):
        if batch_route:
            return {"explanationNumTopClasses": self.num_top_classes}
        return {"numTopClasses": self.num_top_classes}


class ClassListMode(PredictionExplanationsMode):
    """Calculate prediction explanations for the specified classes in each row.

    Attributes
    ----------
    class_names : list
        List of class names that will be explained for each dataset row.
    """

    def __init__(self, class_names):
        self.class_names = class_names

    def get_api_parameters(self, batch_route=False):
        if batch_route:
            return {"explanationClassNames": self.class_names}
        return {"classNames": self.class_names}


class PredictionExplanationsInitialization(APIObject):
    """
    Represents a prediction explanations initialization of a model.

    Attributes
    ----------
    project_id : str
        id of the project the model belongs to
    model_id : str
        id of the model the prediction explanations initialization is for
    prediction_explanations_sample : list of dict
        a small sample of prediction explanations that could be generated for the model
    """

    _path_template = "projects/{}/models/{}/predictionExplanationsInitialization/"
    _converter = t.Dict(
        {
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("prediction_explanations_sample"): t.List(prediction_explanations_trafaret),
        }
    ).allow_extra("*")

    def __init__(self, project_id, model_id, prediction_explanations_sample=None):
        self.project_id = project_id
        self.model_id = model_id
        self.prediction_explanations_sample = prediction_explanations_sample

        self._path = self._path_template.format(self.project_id, self.model_id)

    def __repr__(self):
        return "{}(project_id={}, model_id={})".format(
            type(self).__name__, self.project_id, self.model_id
        )

    @classmethod
    def get(cls, project_id, model_id):
        """
        Retrieve the prediction explanations initialization for a model.

        Prediction explanations initializations are a prerequisite for computing prediction
        explanations, and include a sample what the computed prediction explanations for a
        prediction dataset would look like.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model the prediction explanations initialization is for

        Returns
        -------
        prediction_explanations_initialization : PredictionExplanationsInitialization
            The queried instance.

        Raises
        ------
        ClientError
            If the project or model does not exist or the initialization has not been computed.
        """
        path = cls._path_template.format(project_id, model_id)
        return cls.from_location(path)

    @classmethod
    def create(cls, project_id, model_id):
        """
        Create a prediction explanations initialization for the specified model.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model for which initialization is requested

        Returns
        -------
        job : Job
            an instance of created async job
        """
        from .job import Job  # pylint: disable=import-outside-toplevel,cyclic-import

        response = cls._client.post(cls._path_template.format(project_id, model_id))
        job_id = get_id_from_response(response)
        return Job.get(project_id, job_id)

    def delete(self):
        """
        Delete this prediction explanations initialization.
        """
        self._client.delete(self._path)


class PredictionExplanations(APIObject):
    """
    Represents prediction explanations metadata and provides access to computation results.

    Examples
    --------
    .. code-block:: python

        prediction_explanations = dr.PredictionExplanations.get(project_id, explanations_id)
        for row in prediction_explanations.get_rows():
            print(row)  # row is an instance of PredictionExplanationsRow

    Attributes
    ----------
    id : str
        id of the record and prediction explanations computation result
    project_id : str
        id of the project the model belongs to
    model_id : str
        id of the model the prediction explanations are for
    dataset_id : str
        id of the prediction dataset prediction explanations were computed for
    max_explanations : int
        maximum number of prediction explanations to supply per row of the dataset
    threshold_low : float
        the lower threshold, below which a prediction must score in order for prediction
        explanations to be computed for a row in the dataset
    threshold_high : float
        the high threshold, above which a prediction must score in order for prediction
        explanations to be computed for a row in the dataset
    num_columns : int
        the number of columns prediction explanations were computed for
    finish_time : float
        timestamp referencing when computation for these prediction explanations finished
    prediction_explanations_location : str
        where to retrieve the prediction explanations
    source: str
        For OTV/TS in-training predictions. Holds the portion of the training dataset used to generate
        predictions.
    """

    _path_template = "projects/{}/predictionExplanationsRecords/"
    _expls_path_template = "projects/{}/predictionExplanations/"
    _training_path_template = "projects/{}/predictionExplanationsOnTrainingData/"
    _converter = t.Dict(
        {
            t.Key("id"): String,
            t.Key("project_id"): String,
            t.Key("model_id"): String,
            t.Key("dataset_id"): String,
            t.Key("max_explanations"): Int,
            t.Key("threshold_low", optional=True): t.Float,
            t.Key("threshold_high", optional=True): t.Float,
            t.Key("class_names", optional=True): t.List(t.String),
            t.Key("num_top_classes", optional=True): t.Int(),
            t.Key("num_columns"): Int,
            t.Key("finish_time"): t.Float,
            t.Key("prediction_explanations_location"): String,
            t.Key("source", optional=True): String,
        }
    ).allow_extra("*")

    def __init__(
        self,
        id,
        project_id,
        model_id,
        dataset_id,
        max_explanations,
        num_columns,
        finish_time,
        prediction_explanations_location,
        threshold_low=None,
        threshold_high=None,
        class_names=None,
        num_top_classes=None,
        source=None,
    ):
        self.project_id = project_id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.max_explanations = max_explanations
        self.threshold_low = threshold_low
        self.threshold_high = threshold_high
        self.id = id
        self.num_columns = num_columns
        self.finish_time = datetime.fromtimestamp(finish_time)
        self.prediction_explanations_location = prediction_explanations_location
        self.class_names = class_names
        self.num_top_classes = num_top_classes
        self.source = source

        self._path = self._path_template.format(self.project_id)

    def __repr__(self):
        return "{}(id={}, project_id={}, model_id={})".format(
            type(self).__name__, self.id, self.project_id, self.model_id
        )

    @classmethod
    def get(cls, project_id, prediction_explanations_id):
        """
        Retrieve a specific prediction explanations metadata.

        Parameters
        ----------
        project_id : str
            id of the project the explanations belong to
        prediction_explanations_id : str
            id of the prediction explanations

        Returns
        -------
        prediction_explanations : PredictionExplanations
            The queried instance.
        """
        path = f"{cls._path_template.format(project_id)}{prediction_explanations_id}/"
        return cls.from_location(path)

    @classmethod
    def create(
        cls,
        project_id,
        model_id,
        dataset_id,
        max_explanations=None,
        threshold_low=None,
        threshold_high=None,
        mode=None,
    ):
        """
        Create prediction explanations for the specified dataset.

        In order to create PredictionExplanations for a particular model and dataset, you must
        first:

          * Compute feature impact for the model via ``datarobot.Model.get_feature_impact()``
          * Compute a PredictionExplanationsInitialization for the model via
            ``datarobot.PredictionExplanationsInitialization.create(project_id, model_id)``
          * Compute predictions for the model and dataset via
            ``datarobot.Model.request_predictions(dataset_id)``

        ``threshold_high`` and ``threshold_low`` are optional filters applied to speed up
        computation.  When at least one is specified, only the selected outlier rows will have
        prediction explanations computed. Rows are considered to be outliers if their predicted
        value (in case of regression projects) or probability of being the positive
        class (in case of classification projects) is less than ``threshold_low`` or greater than
        ``thresholdHigh``.  If neither is specified, prediction explanations will be computed for
        all rows.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        model_id : str
            id of the model for which prediction explanations are requested
        dataset_id : str
            id of the prediction dataset for which prediction explanations are requested
        threshold_low : Optional[float]
            the lower threshold, below which a prediction must score in order for prediction
            explanations to be computed for a row in the dataset. If neither ``threshold_high`` nor
            ``threshold_low`` is specified, prediction explanations will be computed for all rows.
        threshold_high : Optional[float]
            the high threshold, above which a prediction must score in order for prediction
            explanations to be computed. If neither ``threshold_high`` nor ``threshold_low`` is
            specified, prediction explanations will be computed for all rows.
        max_explanations : Optional[int]
            the maximum number of prediction explanations to supply per row of the dataset,
            default: 3.
        mode : PredictionExplanationsMode, optional
            mode of calculation for multiclass models, if not specified - server default is
            to explain only the predicted class, identical to passing TopPredictionsMode(1).


        Returns
        -------
        job: Job
            an instance of created async job
        """
        from .job import Job  # pylint: disable=import-outside-toplevel,cyclic-import

        payload = {
            "model_id": model_id,
            "dataset_id": dataset_id,
        }
        if max_explanations is not None:
            payload["max_explanations"] = max_explanations
        if threshold_low is not None:
            payload["threshold_low"] = threshold_low
        if threshold_high is not None:
            payload["threshold_high"] = threshold_high
        if mode is not None:
            payload.update(mode.get_api_parameters(batch_route=False))

        response = cls._client.post(cls._expls_path_template.format(project_id), data=payload)
        job_id = get_id_from_response(response)
        return Job.get(project_id, job_id)

    @classmethod
    def create_on_training_data(
        cls,
        project_id,
        model_id,
        dataset_id,
        max_explanations=None,
        threshold_low=None,
        threshold_high=None,
        mode=None,
        datetime_prediction_partition=None,
    ):
        """
        Create prediction explanations for the the dataset used to train the model.
        This can be retrieved by calling ``dr.Model.get().featurelist_id``.
        For OTV and timeseries projects, ``datetime_prediction_partition`` is required and limited to the
        first backtest ('0') or holdout ('holdout').

        In order to create PredictionExplanations for a particular model and dataset, you must
        first:

          * Compute Feature Impact for the model via ``datarobot.Model.get_feature_impact()``/
          * Compute a PredictionExplanationsInitialization for the model via
            ``datarobot.PredictionExplanationsInitialization.create(project_id, model_id)``.
          * Compute predictions for the model and dataset via
            ``datarobot.Model.request_predictions(dataset_id)``.

        ``threshold_high`` and ``threshold_low`` are optional filters applied to speed up
        computation.  When at least one is specified, only the selected outlier rows will have
        prediction explanations computed. Rows are considered to be outliers if their predicted
        value (in case of regression projects) or probability of being the positive
        class (in case of classification projects) is less than ``threshold_low`` or greater than
        ``thresholdHigh``.  If neither is specified, prediction explanations will be computed for
        all rows.

        Parameters
        ----------
        project_id : str
            The ID of the project the model belongs to.
        model_id : str
            The ID of the model for which prediction explanations are requested.
        dataset_id : str
            The ID of the prediction dataset for which prediction explanations are requested.
        threshold_low : Optional[float]
            The lower threshold, below which a prediction must score in order for prediction
            explanations to be computed for a row in the dataset. If neither ``threshold_high`` nor
            ``threshold_low`` is specified, prediction explanations will be computed for all rows.
        threshold_high : Optional[float]
            The high threshold, above which a prediction must score in order for prediction
            explanations to be computed. If neither ``threshold_high`` nor ``threshold_low`` is
            specified, prediction explanations will be computed for all rows.
        max_explanations : Optional[int]
            The maximum number of prediction explanations to supply per row of the dataset
            (default: 3).
        mode : PredictionExplanationsMode, optional
            The mode of calculation for multiclass models. If not specified, the server default is
            to explain only the predicted class, identical to passing TopPredictionsMode(1).
        datetime_prediction_partition: str
            Options: '0', 'holdout' or None.
            Used only by time series and OTV projects to indicate what part of the dataset
            will be used to generate predictions for computing prediction explanation. Current
            options are '0' (first backtest) and 'holdout'.
            Note that only the validation partition of the first backtest will be used to
            generation predictions.

        Returns
        -------
        job: Job
            An instance of created async job.
        """
        from .job import Job  # pylint: disable=import-outside-toplevel,cyclic-import

        payload = {
            "model_id": model_id,
            "dataset_id": dataset_id,
        }
        if max_explanations is not None:
            payload["max_explanations"] = max_explanations
        if threshold_low is not None:
            payload["threshold_low"] = threshold_low
        if threshold_high is not None:
            payload["threshold_high"] = threshold_high
        if mode is not None:
            payload.update(mode.get_api_parameters(batch_route=False))
        if datetime_prediction_partition is not None:
            payload["datetime_prediction_partition"] = datetime_prediction_partition

        response = cls._client.post(cls._training_path_template.format(project_id), data=payload)
        job_id = get_id_from_response(response)
        return Job.get(project_id, job_id)

    @classmethod
    def list(cls, project_id, model_id=None, limit=None, offset=None):
        """
        List of prediction explanations metadata for a specified project.

        Parameters
        ----------
        project_id : str
            id of the project to list prediction explanations for
        model_id : Optional[str]
            if specified, only prediction explanations computed for this model will be returned
        limit : int or None
            at most this many results are returned, default: no limit
        offset : int or None
            this many results will be skipped, default: 0

        Returns
        -------
        prediction_explanations : list[PredictionExplanations]
        """
        response = cls._client.get(
            cls._path_template.format(project_id),
            params={"model_id": model_id, "limit": limit, "offset": offset},
        )
        r_data = response.json()
        return [cls.from_server_data(item) for item in r_data["data"]]

    def get_rows(self, batch_size=None, exclude_adjusted_predictions=True):
        """
        Retrieve prediction explanations rows.

        Parameters
        ----------
        batch_size : int or None, optional
            maximum number of prediction explanations rows to retrieve per request
        exclude_adjusted_predictions : bool
            Optional, defaults to True. Set to False to include adjusted predictions, which will
            differ from the predictions on some projects, e.g. those with an exposure column
            specified.

        Yields
        ------
        prediction_explanations_row : PredictionExplanationsRow
            Represents prediction explanations computed for a prediction row.
        """
        page = self.get_prediction_explanations_page(
            limit=batch_size, exclude_adjusted_predictions=exclude_adjusted_predictions
        )
        while True:
            for row in page.data:
                yield PredictionExplanationsRow(**row)
            if not page.next_page:
                break
            page = PredictionExplanationsPage.from_location(page.next_page)

    def is_multiclass(self):
        """Whether these explanations are for a multiclass project or a non-multiclass project"""
        # This method will exist now only to maintain backwards compatibility.

        return (self.num_top_classes or self.class_names) and ProjectOptions.get(
            project_id=self.project_id
        ).target == TARGET_TYPE.MULTICLASS

    def is_unsupervised_clustering_or_multiclass(self):
        """Clustering and multiclass XEMP always has either one of num_top_classes or class_names
        parameters set"""
        return bool(self.num_top_classes or self.class_names)

    def get_number_of_explained_classes(self):
        """How many classes we attempt to explain for each row"""
        if self.is_unsupervised_clustering_or_multiclass():
            return self.num_top_classes or len(self.class_names)
        return 1

    def get_all_as_dataframe(self, exclude_adjusted_predictions=True):
        """
        Retrieve all prediction explanations rows and return them as a pandas.DataFrame.

        Returned dataframe has the following structure:

            - row_id : row id from prediction dataset
            - prediction : the output of the model for this row
            - adjusted_prediction : adjusted prediction values (only appears for projects that
              utilize prediction adjustments, e.g. projects with an exposure column)
            - class_0_label : a class level from the target (only appears for classification
              projects)
            - class_0_probability : the probability that the target is this class (only appears for
              classification projects)
            - class_1_label : a class level from the target (only appears for classification
              projects)
            - class_1_probability : the probability that the target is this class (only appears for
              classification projects)
            - explanation_0_feature : the name of the feature contributing to the prediction for
              this explanation
            - explanation_0_feature_value : the value the feature took on
            - explanation_0_label : the output being driven by this explanation.  For regression
              projects, this is the name of the target feature.  For classification projects, this
              is the class label whose probability increasing would correspond to a positive
              strength.
            - explanation_0_qualitative_strength : a human-readable description of how strongly the
              feature affected the prediction (e.g. '+++', '--', '+') for this explanation
            - explanation_0_per_ngram_text_explanations : Text prediction explanations data in json
              formatted string.
            - explanation_0_strength : the amount this feature's value affected the prediction
            - ...
            - explanation_N_feature : the name of the feature contributing to the prediction for
              this explanation
            - explanation_N_feature_value : the value the feature took on
            - explanation_N_label : the output being driven by this explanation.  For regression
              projects, this is the name of the target feature.  For classification projects, this
              is the class label whose probability increasing would correspond to a positive
              strength.
            - explanation_N_qualitative_strength : a human-readable description of how strongly the
              feature affected the prediction (e.g. '+++', '--', '+') for this explanation
            - explanation_N_per_ngram_text_explanations : Text prediction explanations data in json
              formatted string.
            - explanation_N_strength : the amount this feature's value affected the prediction

        For classification projects, the server does not guarantee any ordering on the prediction
        values, however within this function we sort the values so that `class_X` corresponds to
        the same class from row to row.

        Parameters
        ----------
        exclude_adjusted_predictions : bool
            Optional, defaults to True. Set this to False to include adjusted prediction values in
            the returned dataframe.

        Returns
        -------
        dataframe: pandas.DataFrame
        """
        columns = ["row_id", "prediction"]
        rows = self.get_rows(
            batch_size=1, exclude_adjusted_predictions=exclude_adjusted_predictions
        )
        first_row = next(rows)
        has_text_explanations = (
            first_row.prediction_explanations
            and "per_ngram_text_explanations" in first_row.prediction_explanations[0]
        )
        adjusted_predictions_in_data = first_row.adjusted_prediction is not None
        if adjusted_predictions_in_data:
            columns.append("adjusted_prediction")
        # for regression, length is 1; for classification, length is number of levels in target
        # i.e. 2 for binary classification
        is_classification = len(first_row.prediction_values) > 1
        # include class label/probability for classification project
        if is_classification:
            for i in range(len(first_row.prediction_values)):
                columns.extend([f"class_{i}_label", f"class_{i}_probability"])
        if self.is_unsupervised_clustering_or_multiclass():
            for class_num in range(self.get_number_of_explained_classes()):
                prefix = f"explained_class_{class_num}"
                for i in range(self.max_explanations):
                    columns.extend(
                        [
                            f"{prefix}_explanation_{i}_feature",
                            f"{prefix}_explanation_{i}_feature_value",
                            f"{prefix}_explanation_{i}_label",
                            f"{prefix}_explanation_{i}_qualitative_strength",
                            f"{prefix}_explanation_{i}_strength",
                        ]
                    )
                    if has_text_explanations:
                        columns.append(f"{prefix}_explanation_{i}_per_ngram_text_explanations")

        else:
            for i in range(self.max_explanations):
                columns.extend(
                    [
                        f"explanation_{i}_feature",
                        f"explanation_{i}_feature_value",
                        f"explanation_{i}_label",
                        f"explanation_{i}_qualitative_strength",
                        f"explanation_{i}_strength",
                    ]
                )
                if has_text_explanations:
                    columns.append(f"explanation_{i}_per_ngram_text_explanations")
        pred_expl_list = []

        for i, row in enumerate(
            self.get_rows(exclude_adjusted_predictions=exclude_adjusted_predictions)
        ):
            data = [row.row_id, row.prediction]
            if adjusted_predictions_in_data:
                data.append(row.adjusted_prediction)
            if is_classification:
                for pred_value in sorted(row.prediction_values, key=lambda x: x["label"]):
                    data.extend([pred_value["label"], pred_value["value"]])
            if self.is_unsupervised_clustering_or_multiclass():
                # Each explained class can have less then self.max_explanations and needs to be
                # padded with None separately
                current_label = ""
                explanations_per_label = self.max_explanations
                for pred_expl in row.prediction_explanations:
                    if pred_expl["label"] == current_label:
                        explanations_per_label += 1
                    else:
                        # We switched to explaining next label, check if padding is needed
                        if explanations_per_label < self.max_explanations:
                            data += [None] * (self.max_explanations - explanations_per_label) * 5
                        current_label = pred_expl["label"]
                        explanations_per_label = 1
                    data.extend(
                        [
                            pred_expl["feature"],
                            pred_expl["feature_value"],
                            pred_expl["label"],
                            pred_expl["qualitative_strength"],
                            pred_expl["strength"],
                        ]
                    )
                    if has_text_explanations:
                        data.append(
                            json.dumps(
                                pred_expl.get("per_ngram_text_explanations"), ensure_ascii=False
                            )
                        )
                pred_expl_list.append(data + [None] * (len(columns) - len(data)))
            else:
                for pred_expl in row.prediction_explanations:
                    data.extend(
                        [
                            pred_expl["feature"],
                            pred_expl["feature_value"],
                            pred_expl["label"],
                            pred_expl["qualitative_strength"],
                            pred_expl["strength"],
                        ]
                    )
                    if has_text_explanations:
                        data.append(
                            json.dumps(
                                pred_expl.get("per_ngram_text_explanations"), ensure_ascii=False
                            )
                        )
                pred_expl_list.append(data + [None] * (len(columns) - len(data)))

        return pd.DataFrame(data=pred_expl_list, columns=columns)

    def download_to_csv(self, filename, encoding="utf-8", exclude_adjusted_predictions=True):
        """
        Save prediction explanations rows into CSV file.

        Parameters
        ----------
        filename : str or file object
            path or file object to save prediction explanations rows
        encoding : string, optional
            A string representing the encoding to use in the output file, defaults to 'utf-8'
        exclude_adjusted_predictions : bool
            Optional, defaults to True. Set to False to include adjusted predictions, which will
            differ from the predictions on some projects, e.g. those with an exposure column
            specified.
        """
        df = self.get_all_as_dataframe(exclude_adjusted_predictions=exclude_adjusted_predictions)
        df.to_csv(path_or_buf=filename, header=True, index=False, encoding=encoding)

    def get_prediction_explanations_page(
        self, limit=None, offset=None, exclude_adjusted_predictions=True
    ):
        """
        Get prediction explanations.

        If you don't want use a generator interface, you can access paginated prediction
        explanations directly.

        Parameters
        ----------
        limit : int or None
            the number of records to return, the server will use a (possibly finite) default if not
            specified
        offset : int or None
            the number of records to skip, default 0
        exclude_adjusted_predictions : bool
            Optional, defaults to True. Set to False to include adjusted predictions, which will
            differ from the predictions on some projects, e.g. those with an exposure column
            specified.

        Returns
        -------
        prediction_explanations : PredictionExplanationsPage
        """
        kwargs = {"limit": limit, "exclude_adjusted_predictions": exclude_adjusted_predictions}
        if offset:
            kwargs["offset"] = offset
        return PredictionExplanationsPage.get(self.project_id, self.id, **kwargs)

    def delete(self):
        """
        Delete these prediction explanations.
        """
        path = f"{self._path_template.format(self.project_id)}{self.id}/"
        self._client.delete(path)


class PredictionExplanationsRow:
    """
    Represents prediction explanations computed for a prediction row.

    Notes
    -----

    ``PredictionValue`` contains:

    * ``label`` : describes what this model output corresponds to.  For regression projects,
      it is the name of the target feature.  For classification projects, it is a level from
      the target feature.
    * ``value`` : the output of the prediction.  For regression projects, it is the predicted
      value of the target.  For classification projects, it is the predicted probability the
      row belongs to the class identified by the label.


    ``PredictionExplanation`` contains:

    * ``label`` : described what output was driven by this explanation.  For regression
      projects, it is the name of the target feature.  For classification projects, it is the
      class whose probability increasing would correspond to a positive strength of this
      prediction explanation.
    * ``feature`` : the name of the feature contributing to the prediction
    * ``feature_value`` : the value the feature took on for this row
    * ``strength`` : the amount this feature's value affected the prediction
    * ``qualitative_strength`` : a human-readable description of how strongly the feature
      affected the prediction. A large positive effect is denoted '+++', medium '++', small '+',
      very small '<+'. A large negative effect is denoted '---', medium '--', small '-', very
      small '<-'.

    Attributes
    ----------
    row_id : int
        which row this ``PredictionExplanationsRow`` describes
    prediction : float
        the output of the model for this row
    adjusted_prediction : float or None
        adjusted prediction value for projects that provide this information, None otherwise
    prediction_values : list
        an array of dictionaries with a schema described as ``PredictionValue``
    adjusted_prediction_values : list
        same as prediction_values but for adjusted predictions
    prediction_explanations : list
        an array of dictionaries with a schema described as ``PredictionExplanation``
    """

    def __init__(
        self,
        row_id,
        prediction,
        prediction_values,
        prediction_explanations=None,
        adjusted_prediction=None,
        adjusted_prediction_values=None,
    ):
        self.row_id = row_id
        self.prediction = prediction
        self.prediction_values = prediction_values
        self.prediction_explanations = prediction_explanations
        self.adjusted_prediction = adjusted_prediction
        self.adjusted_prediction_values = adjusted_prediction_values

    def __repr__(self):
        return "{}(row_id={}, prediction={})".format(
            type(self).__name__, self.row_id, self.prediction
        )


class PredictionExplanationsPage(APIObject):
    """
    Represents a batch of prediction explanations received by one request.

    Attributes
    ----------
    id : str
        id of the prediction explanations computation result
    data : list[dict]
        list of raw prediction explanations; each row corresponds to a row of the prediction dataset
    count : int
        total number of rows computed
    previous_page : str
        where to retrieve previous page of prediction explanations, None if current page is the
        first
    next_page : str
        where to retrieve next page of prediction explanations, None if current page is the last
    prediction_explanations_record_location : str
        where to retrieve the prediction explanations metadata
    adjustment_method : str
        Adjustment method that was applied to predictions, or 'N/A' if no adjustments were done.
    """

    _path_template = "projects/{}/predictionExplanations/"
    _converter = t.Dict(
        {
            t.Key("id"): String,
            t.Key("count"): Int,
            t.Key("previous", optional=True): String(),
            t.Key("next", optional=True): String(),
            t.Key("data"): t.List(prediction_explanations_trafaret),
            t.Key("prediction_explanations_record_location"): t.URL,
            t.Key("adjustment_method", default="N/A"): String(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        id,
        count=None,
        previous=None,
        next=None,
        data=None,
        prediction_explanations_record_location=None,
        adjustment_method=None,
    ):
        self.id = id
        self.count = count
        self.previous_page = previous
        self.next_page = next
        self.data = data
        self.prediction_explanations_record_location = prediction_explanations_record_location
        self.adjustment_method = adjustment_method

    def __repr__(self):
        return f"{type(self).__name__}(id={self.id})"

    @classmethod
    def get(
        cls,
        project_id,
        prediction_explanations_id,
        limit=None,
        offset=0,
        exclude_adjusted_predictions=True,
    ):
        """
        Retrieve prediction explanations.

        Parameters
        ----------
        project_id : str
            id of the project the model belongs to
        prediction_explanations_id : str
            id of the prediction explanations
        limit : int or None
            the number of records to return; the server will use a (possibly finite) default if not
            specified
        offset : int or None
            the number of records to skip, default 0
        exclude_adjusted_predictions : bool
            Optional, defaults to True. Set to False to include adjusted predictions, which will
            differ from the predictions on some projects, e.g. those with an exposure column
            specified.

        Returns
        -------
        prediction_explanations : PredictionExplanationsPage
            The queried instance.
        """
        params = {
            "offset": offset,
            "exclude_adjusted_predictions": "true" if exclude_adjusted_predictions else "false",
        }
        if limit:
            params["limit"] = limit
        path = f"{cls._path_template.format(project_id)}{prediction_explanations_id}/"
        return cls.from_location(path, keep_attrs=KEEP_ATTRS, params=params)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Optional, TYPE_CHECKING

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    # We are using the "non-inheritance" instantiation because if trying
    # `class WordCloudNgram(TypedDict):` we have a syntax error thanks to the use of the reserved
    # word `class` - this "namedtuple-like" approach to instantiation gets around that issue
    WordCloudNgram = TypedDict(
        "WordCloudNgram",
        {
            "ngram": str,
            "coefficient": float,
            "count": int,
            "frequency": float,
            "is_stopword": bool,
            "class": Optional[str],
            "variable": Optional[str],
        },
    )


class WordCloud(APIObject):
    """ Word cloud data for the model.

    Notes
    -----
    ``WordCloudNgram`` is a dict containing the following:

        * ``ngram`` (str) Word or ngram value.
        * ``coefficient`` (float) Value from [-1.0, 1.0] range, describes effect of this ngram on \
          the target. Large negative value means strong effect toward negative class in \
          classification and smaller target value in regression models. Large positive - toward \
          positive class and bigger value respectively.
        * ``count`` (int) Number of rows in the training sample where this ngram appears.
        * ``frequency`` (float) Value from (0.0, 1.0] range, relative frequency of given ngram to \
          most frequent ngram.
        * ``is_stopword`` (bool) True for ngrams that DataRobot evaluates as stopwords.
        * ``class`` (str or None) For classification - values of the target class for
          corresponding word or ngram. For regression - None.

    Attributes
    ----------
    ngrams : list of dict
        List of dicts with schema described as ``WordCloudNgram`` above.
    """

    _converter = t.Dict(
        {
            t.Key("ngrams"): t.List(
                t.Dict(
                    {
                        t.Key("ngram"): String,
                        t.Key("coefficient"): t.Float(gte=-1, lte=1),
                        t.Key("count"): Int,
                        t.Key("frequency"): t.Float(gt=0, lte=1),
                        t.Key("is_stopword"): t.Bool,
                        # Making these optional will allow working with older backends
                        t.Key("class", optional=True, default=None): t.Or(t.String, t.Null),
                        t.Key("variable", optional=True, default=None): t.Or(t.String, t.Null),
                    }
                ).ignore_extra("*")
            )
        }
    ).ignore_extra("*")

    def __init__(self, ngrams: List[WordCloudNgram]) -> None:
        self.ngrams = ngrams

    def __repr__(self) -> str:
        return f"WordCloud({len(self.ngrams)} ngrams)"

    def most_frequent(self, top_n: Optional[int] = 5) -> List[WordCloudNgram]:
        """Return most frequent ngrams in the word cloud.

        Parameters
        ----------
        top_n : int
            Number of ngrams to return

        Returns
        -------
        list of dict
            Up to top_n top most frequent ngrams in the word cloud.
            If top_n bigger then total number of ngrams in word cloud - return all sorted by
            frequency in descending order.
        """
        return sorted(self.ngrams, key=lambda ngram: ngram["frequency"], reverse=True)[:top_n]

    def most_important(self, top_n: Optional[int] = 5) -> List[WordCloudNgram]:
        """Return most important ngrams in the word cloud.

        Parameters
        ----------
        top_n : int
            Number of ngrams to return

        Returns
        -------
        list of dict
            Up to top_n top most important ngrams in the word cloud.
            If top_n bigger then total number of ngrams in word cloud - return all sorted by
            absolute coefficient value in descending order.
        """
        return sorted(self.ngrams, key=lambda ngram: abs(ngram["coefficient"]), reverse=True)[
            :top_n
        ]

    def ngrams_per_class(self) -> Dict[Optional[str], List[WordCloudNgram]]:
        """Split ngrams per target class values. Useful for multiclass models.

        Returns
        -------
        dict
            Dictionary in the format of (class label) -> (list of ngrams for that class)
        """
        per_class = defaultdict(list)
        for ngram in self.ngrams:
            per_class[ngram["class"]].append(ngram)
        return dict(per_class)



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import trafaret as t

from datarobot.models.api_object import APIObject


class UserLimits(APIObject):
    """
    Counts for user limits for LLM APIs and vector databases.
    """

    _llm_path = "api/v2/genai/userLimits/llmApiCalls/"
    _vector_db_path = "api/v2/genai/userLimits/vectorDatabases/"

    _converter = t.Dict(
        {
            t.Key("counter"): t.Int(),
        }
    ).ignore_extra("*")

    def __init__(self, counter: int):
        self.counter = counter

    @classmethod
    def get_vector_database_count(cls) -> APIObject:
        """Get the count of vector databases for the user."""
        url = f"{cls._client.domain}/{cls._vector_db_path}"
        return cls.from_location(url)

    @classmethod
    def get_llm_requests_count(cls) -> APIObject:
        """Get the count of LLMs requests made by the user."""
        url = f"{cls._client.domain}/{cls._llm_path}"
        return cls.from_location(url)



"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from datetime import date, datetime
import re
from typing import Any, cast, Dict, Iterable, List, Match, Optional, Tuple, TYPE_CHECKING, Union

from dateutil import parser, tz
import numpy as np
import pandas as pd
import pytz

from .deprecation import deprecated, deprecation_warning  # noqa
from .sourcedata import dataframe_to_buffer, is_urlsource, recognize_sourcedata  # noqa

ALL_CAPITAL = re.compile(r"(.)([A-Z][a-z]+)")
CASE_SWITCH = re.compile(r"([a-z0-9])([A-Z])")
UNDERSCORES = re.compile(r"([a-z]?)(_+)([a-z])")

if TYPE_CHECKING:
    from requests import Response

    from datarobot.models.api_object import (
        APIObject,
        ServerDataDictType,
        ServerDataListType,
        ServerDataType,
    )
    from datarobot.models.predictions import PredictionsResponse


ApiItemType = Union[
    Dict[str, Any],
    "APIObject",
    datetime,
    date,
]
ToApiReturnType = Union[str, Dict[str, Any], List[Dict[str, Any]]]


class rawdict(dict):  # type: ignore[type-arg]
    """
    Dictionaries returned from models will have their keys snakeCased.
    Wrapping them in rawdict will pass them through to the API verbatim.
    """


def underscorize(value: str) -> str:
    partial_result = ALL_CAPITAL.sub(r"\1_\2", value)
    return CASE_SWITCH.sub(r"\1_\2", partial_result).lower()


def underscoreToCamel(match: Match[str]) -> str:
    prefix, underscores, postfix = match.groups()
    if len(underscores) > 1:
        # underscoreToCamel('sample_pct__gte') -> 'samplePct__gte'
        return match.group()
    return prefix + postfix.upper()


def camelize(value: str) -> str:
    return UNDERSCORES.sub(underscoreToCamel, value)


# pylint: disable-next=missing-function-docstring
def from_api(
    data: ServerDataType,
    do_recursive: bool = True,
    keep_attrs: Optional[Union[Iterable[str], Iterable[List[str]]]] = None,
    keep_null_keys: bool = False,
) -> ServerDataType:
    if type(data) not in (dict, list):
        return data
    if isinstance(data, list):
        return _from_api_list(
            data, do_recursive=do_recursive, keep_attrs=keep_attrs, keep_null_keys=keep_null_keys
        )
    return _from_api_dict(
        data, do_recursive=do_recursive, keep_attrs=keep_attrs, keep_null_keys=keep_null_keys
    )


# pylint: disable-next=missing-function-docstring
def _from_api_dict(
    data: ServerDataDictType,
    do_recursive: bool = True,
    keep_attrs: Optional[Union[Iterable[str], Iterable[List[str]]]] = None,
    keep_null_keys: bool = False,
) -> Dict[str, Any]:
    keep_attrs = keep_attrs or []
    # prepare attributes in format 'top.middle.bottom' for processing
    parsed_attrs = []
    for attr in keep_attrs:
        if isinstance(attr, str):
            parsed_attrs.append(attr.split("."))
        else:
            parsed_attrs.append(attr)
    # take index 0 since recursion goes from top to bottom
    current_level = [attr.pop(0) for attr in parsed_attrs if len(attr)]
    # filter out empty attrs to pass through recursive call
    next_level_attrs = [attr for attr in parsed_attrs if attr]

    app_data = {}
    for k, v in data.items():
        k_under = underscorize(k)
        if v is None and k_under not in current_level and not keep_null_keys:
            continue
        if do_recursive:
            data_val = from_api(
                v,
                do_recursive=do_recursive,
                keep_attrs=deepcopy(next_level_attrs),
                keep_null_keys=keep_null_keys,
            )
        else:
            data_val = v
        app_data[k_under] = data_val
    return app_data


def _from_api_list(
    data: ServerDataListType,
    do_recursive: bool = True,
    keep_attrs: Optional[Union[Iterable[str], Iterable[List[str]]]] = None,
    keep_null_keys: bool = False,
) -> List[Any]:
    return [
        from_api(
            datum,
            do_recursive=do_recursive,
            keep_attrs=deepcopy(keep_attrs),
            keep_null_keys=keep_null_keys,
        )
        for datum in data
    ]


def remove_empty_keys(
    metadata: Dict[str, Any],
    keep_attrs: Optional[List[str]] = None,
) -> Dict[str, Any]:
    keep_attrs = keep_attrs or []
    if any(not isinstance(k_a, str) for k_a in keep_attrs):
        raise ValueError("All items in keep_attrs supplied must be strings")
    return {k: v for k, v in metadata.items() if v is not None or k in keep_attrs}


def assert_single_parameter(param_names: Iterable[str], *params: Any) -> None:
    if sum(param is not None for param in params) != 1:
        raise TypeError(f"One and only parameter of: {param_names}")


def assert_single_or_zero_parameter(param_names: Iterable[str], *params: Any) -> None:
    if sum(param is not None for param in params) > 1:
        raise TypeError(f"Either none or one and only parameter of: {param_names}")


def parse_time(time_str: str) -> Union[str, datetime]:
    try:
        return parser.parse(time_str, tzinfos={"UTC": tz.tzutc()})
    except Exception:
        return time_str


def datetime_to_string(
    datetime_obj: Any,
    ensure_rfc_3339: bool = False,
) -> str:
    """Converts to isoformat"""
    if not isinstance(datetime_obj, datetime):
        msg = f"expected to be passed a datetime.datetime, was passed {type(datetime_obj)}"
        raise ValueError(msg)
    if ensure_rfc_3339 and not datetime_obj.tzinfo:
        datetime_obj = datetime_obj.replace(tzinfo=pytz.utc)
    return datetime_obj.isoformat()


def _is_api_object(obj: Any) -> bool:
    # local import of APIObject is required due to circular imports between
    # utils.__init__.py, rest.py, client.py, and api_object.py if APIObject
    # is imported at the top of the file.
    from datarobot.models.api_object import (  # pylint: disable=import-outside-toplevel,cyclic-import
        APIObject,
    )

    return issubclass(type(obj), APIObject)


def camelize_obj(data: Any) -> Any:
    """Camelize nested object keys"""
    if not data:
        return data
    if isinstance(data, rawdict):
        return data
    elif isinstance(data, dict):
        return {camelize(k): camelize_obj(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [camelize_obj(subitem) for subitem in data]
    else:
        return data


def to_api(
    data: Any,
    keep_attrs: Optional[Union[List[str], List[List[str]]]] = None,
) -> ToApiReturnType:
    """
    :param data: dictionary {'max_digits': 1}
    :return: {'maxDigits': 1}
    """
    if not data:
        return {}
    assert isinstance(data, dict) or _is_api_object(data), "Wrong type"
    return _to_api_item(data, keep_attrs)


def _to_dict(item: Any) -> Dict[str, Any]:
    return {k: v for k, v in item.__dict__.items() if not k.startswith("_") and not callable(v)}


# pylint: disable-next=missing-function-docstring
def _to_api_item(
    item: Union[ApiItemType, List[ApiItemType]],
    keep_attrs: Optional[Union[List[str], List[List[str]]]] = None,
) -> ToApiReturnType:
    if isinstance(item, rawdict):
        return item
    if _is_api_object(item):
        return _to_api_item(item=_to_dict(item), keep_attrs=keep_attrs)
    elif isinstance(item, dict):
        if any(not isinstance(k_a, str) for k_a in keep_attrs or []):
            raise ValueError("All items in keep_attrs supplied must be strings")
        # Casting because we confirm above that we only have a list of strings or None
        dense_item = remove_empty_keys(item, cast(Optional[List[str]], keep_attrs))
        return {camelize(k): _to_api_item(v, keep_attrs=keep_attrs) for k, v in dense_item.items()}
    elif isinstance(item, list):
        # Casting because of misc error seemingly caused by recursion:
        # List comprehension has incompatible type
        return cast(
            ToApiReturnType, [_to_api_item(subitem, keep_attrs=keep_attrs) for subitem in item]
        )
    elif isinstance(item, datetime):
        return datetime_to_string(item)
    elif isinstance(item, date):
        return item.isoformat()
    else:
        # Casting here in part because mypy doesn't understand that `if _is_api_object(item):`
        # above takes care of the APIObject case - so it believes there's a chance that's what we
        # may be returning here
        return cast(ToApiReturnType, item)


def get_id_from_response(response: Response) -> str:
    location_string = response.headers["Location"]
    return get_id_from_location(location_string)


def get_id_from_location(location_string: str) -> str:
    return location_string.split("/")[-2]


def get_duplicate_features(features: List[str]) -> List[str]:
    duplicate_features = set()
    seen_features = set()
    for feature in features:
        if feature in seen_features:
            duplicate_features.add(feature)
        else:
            seen_features.add(feature)
    return list(duplicate_features)


def raw_prediction_response_to_dataframe(
    pred_response: PredictionsResponse,
    class_prefix: str,
) -> pd.DataFrame:
    """Raw predictions for classification come as nested json objects.

    This will extract it to be a single dataframe.

    Parameters
    ----------
    pred_response : dict
        The loaded json object returned from the prediction route.
    class_prefix : str
            The prefix to append to labels in the final dataframe (e.g., apple -> class_apple)

    Returns
    -------
    frame : pandas.DataFrame

    """
    # warning: we can't use the `from_api` function here due to the performance issues when
    # converting larger prediction responses
    predictions = pred_response["predictions"]
    task_type = pred_response["task"]

    frame = pd.DataFrame.from_records(predictions)
    frame = _underscorize_df_columns(frame)
    if task_type == "Multilabel":
        frame = _pivot_multilabel_predictions(frame, class_prefix)
    else:
        has_prediction_values = "prediction_values" in frame.columns
        if has_prediction_values:
            frame = _pivot_prediction_labels(frame, class_prefix)
        if "prediction_explanations" in frame.columns:
            frame = _pivot_prediction_explanations(frame, pred_response, has_prediction_values)

    # Remove some columns with all null values
    # (i.e. TS models return empty series_id for single-series projects)
    drop_cols = [
        col
        for col in ("positive_probability", "series_id")
        if col in frame and frame[col].isna().all()
    ]
    if drop_cols:
        frame = frame.drop(columns=drop_cols)

    return frame


def _underscorize_df_columns(frame: pd.DataFrame) -> pd.DataFrame:
    """Rename the DataFrame columns from the camel case to underscore

    Parameters
    ----------
    frame : pandas.DataFrame

    Returns
    -------
    frame : pandas.DataFrame

    """
    original_columns = list(frame.columns)
    converted_columns = [underscorize(col) for col in original_columns]
    column_mapping = dict(zip(original_columns, converted_columns))
    return frame.rename(columns=column_mapping)


def _pivot_multilabel_predictions(
    frame: pd.DataFrame,
    class_prefix: str,
) -> pd.DataFrame:
    """
    Explodes the prediction_values column into multiple columns for prediction labels, thresholds,
    and predictions.

    Parameters
    ----------
    frame : pandas.DataFrame
        The initial dataframe obtained from prediction response.
    class_prefix : str
        The prefix to append to labels in the final dataframe (e.g., apple -> class_apple)

    Returns
    -------
    pandas.DataFrame
        The converted predictions
    """
    if "prediction_values" not in frame:
        return frame
    labels = [entry["label"] for entry in frame.prediction_values[0]]

    values = np.empty((frame.shape[0], len(labels)), np.float32)
    thresholds = np.empty((frame.shape[0], len(labels)), np.float32)
    predictions = np.empty((frame.shape[0], len(labels)), np.int8)
    for i, pair in enumerate(zip(frame.prediction_values, frame.prediction)):
        row, prediction = pair
        values[i] = np.array([[col["value"] for col in row]])
        thresholds[i] = np.array([[col["threshold"] for col in row]])
        predictions[i] = np.array([[1 if col["label"] in prediction else 0 for col in row]])

    values_df = pd.DataFrame(values, columns=[f"{class_prefix}{label}" for label in labels])
    thresholds_df = pd.DataFrame(thresholds, columns=[f"threshold_{label}" for label in labels])
    predictions_df = pd.DataFrame(predictions, columns=[f"prediction_{label}" for label in labels])
    return pd.concat(
        [
            frame.drop(["prediction_values", "prediction"], axis=1),
            values_df,
            thresholds_df,
            predictions_df,
        ],
        axis=1,
    )


def _pivot_prediction_explanations(
    frame: pd.DataFrame,
    pred_response: PredictionsResponse,
    has_prediction_values: bool,
) -> pd.DataFrame:
    """Pivot the prediction_explanations from the dictionary to columns

    frame : DataFrame
        the DataFrame build based on predictions response
    pred_response : dict
        the response that contains explanations details applicable to all predictions
    has_prediction_values : bool
        flag showing whether there are prediction values for multiple class
    Returns
    -------
    frame : DataFrame
         the dataframe with parsed prediction explanations
    """
    _FEATURE = "Explanation_{}_feature_name"
    _FEATURE_VAL = "Explanation_{}_feature_value"
    _STRENGTH = "Explanation_{}_strength"

    columns: Tuple[str, ...] = ()
    if has_prediction_values:
        columns += ("explained_class",)
    first_row = frame["prediction_explanations"][0]
    for i in range(len(first_row)):
        idx = i + 1
        columns += (
            _FEATURE.format(idx),
            _FEATURE_VAL.format(idx),
            _STRENGTH.format(idx),
        )
    data = []
    for prediction_explanations in frame["prediction_explanations"]:
        data_row: Tuple[Any, ...] = ()
        if has_prediction_values:
            data_row += (prediction_explanations[0]["label"],)
        for prediction_explanation in prediction_explanations:
            data_row += (
                prediction_explanation["feature"],
                prediction_explanation["featureValue"],
                prediction_explanation["strength"],
            )
        data.append(data_row)
    prediction_explanation_df = pd.DataFrame.from_records(data, columns=columns)
    frame = pd.concat([frame, prediction_explanation_df], axis=1)
    frame = frame.drop("prediction_explanations", axis=1)

    frame["shap_remaining_total"] = frame["prediction_explanation_metadata"].apply(
        lambda x: x.get("shapRemainingTotal", None)
    )
    frame = frame.drop("prediction_explanation_metadata", axis=1)

    # shapBaseValue is either list with single float item or scalar float
    shap_base_value = pred_response["shapBaseValue"]
    if isinstance(shap_base_value, list):
        shap_base_value = shap_base_value[0]  # type: ignore[assignment]
    frame["shap_base_value"] = shap_base_value
    return frame


def _pivot_prediction_labels(
    frame: pd.DataFrame,
    class_prefix: str,
) -> pd.DataFrame:
    """Pivot the prediciton_values from the dictionary to columns

    Parameters
    ----------
    frame : pandas.DataFrame

    Returns
    -------
    frame : pandas.DataFrame

    """
    wrapper = defaultdict(list)
    for pred_row in frame["prediction_values"]:
        for pred_value in pred_row:
            col_name = "".join((class_prefix, "{}".format(pred_value["label"])))
            wrapper[col_name].append(pred_value["value"])
    pred_frame = pd.DataFrame.from_records(wrapper)
    frame = pd.concat([frame, pred_frame], axis=1)
    return frame.drop("prediction_values", axis=1)



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
import json
import os
from typing import Dict, Iterable, List, Optional, TYPE_CHECKING, Union

from requests import Response
import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.models.sharing import SharingAccess
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

from .. import errors

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class CountryCode(TypedDict):
        name: str
        code: str


class CalendarFile(APIObject):
    """Represents the data for a calendar file.

    For more information about calendar files, see the
    :ref:`calendar documentation <calendar-files>`.

    Attributes
    ----------
    id : str
        The id of the calendar file.
    calendar_start_date : str
        The earliest date in the calendar.
    calendar_end_date : str
        The last date in the calendar.
    created  : str
        The date this calendar was created, i.e. uploaded to DR.
    name : str
        The name of the calendar.
    num_event_types : int
        The number of different event types.
    num_events : int
        The number of events this calendar has.
    project_ids : list of strings
        A list containing the projectIds of the projects using this calendar.
    multiseries_id_columns: List[str] or None
        A list of columns in calendar which uniquely identify events for different series.
        Currently, only one column is supported.
        If multiseries id columns are not provided, calendar is considered to be single series.
    role : str
        The access role the user has for this calendar.
    """

    _base_url = "calendars/"

    _upload_url = _base_url + "fileUpload/"
    _calendar_url = _base_url + "{}/"
    _access_control_url = _calendar_url + "accessControl/"
    _from_country_code_url = _base_url + "fromCountryCode/"
    _from_dataset_url = _base_url + "fromDataset/"
    _allowed_countries_list_url = "calendarCountryCodes/"

    _converter = t.Dict(
        {
            t.Key("calendar_end_date"): String,
            t.Key("calendar_start_date"): String,
            t.Key("created"): String,
            t.Key("id"): String,
            t.Key("name"): String,
            t.Key("num_event_types"): Int,
            t.Key("num_events"): Int,
            t.Key("project_ids"): t.List(String),
            t.Key("role"): String,
            t.Key("multiseries_id_columns", optional=True): t.Or(t.List(String), t.Null),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        calendar_end_date: Optional[str] = None,
        calendar_start_date: Optional[str] = None,
        created: Optional[str] = None,
        id: Optional[str] = None,
        name: Optional[str] = None,
        num_event_types: Optional[int] = None,
        num_events: Optional[int] = None,
        project_ids: Optional[List[str]] = None,
        role: Optional[str] = None,
        multiseries_id_columns: Optional[List[str]] = None,
    ) -> None:
        self.calendar_end_date = calendar_end_date
        self.calendar_start_date = calendar_start_date
        self.created = created
        self.id = id
        self.name = name
        self.num_event_types = num_event_types
        self.num_events = num_events
        self.project_ids = project_ids
        self.role = role
        self.multiseries_id_columns = multiseries_id_columns

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id})"

    def __eq__(self, other: CalendarFile) -> bool:  # type: ignore[override]
        """Compares the relevant fields of two calendars to assess equality"""
        vars1 = [self.__getattribute__(x) for x in [k.name for k in self._converter.keys]]
        vars2 = [other.__getattribute__(x) for x in [k.name for k in other._converter.keys]]
        return vars1 == vars2

    @classmethod
    def create(
        cls,
        file_path: str,
        calendar_name: Optional[str] = None,
        multiseries_id_columns: Optional[List[str]] = None,
    ) -> CalendarFile:
        """
        Creates a calendar using the given file. For information about calendar files, see the
        :ref:`calendar documentation <calendar-files>`

        The provided file must be a CSV in the format:

        .. code-block:: text

            Date,   Event,          Series ID,    Event Duration
            <date>, <event_type>,   <series id>,  <event duration>
            <date>, <event_type>,              ,  <event duration>

        A header row is required, and the "Series ID" and "Event Duration" columns are optional.

        Once the CalendarFile has been created, pass its ID with
        the :class:`DatetimePartitioningSpecification <datarobot.DatetimePartitioningSpecification>`
        when setting the target for a time series project in order to use it.

        Parameters
        ----------
        file_path : string
            A string representing a path to a local csv file.
        calendar_name : string, optional
            A name to assign to the calendar. Defaults to the name of the file if not provided.
        multiseries_id_columns : List[str] or None
            A list of the names of multiseries id columns to define which series an event
            belongs to. Currently only one multiseries id column is supported.

        Returns
        -------
        calendar_file : CalendarFile
            Instance with initialized data.

        Raises
        ------
        AsyncProcessUnsuccessfulError
            Raised if there was an error processing the provided calendar file.

        Examples
        --------
        .. code-block:: python

            # Creating a calendar with a specified name
            cal = dr.CalendarFile.create('/home/calendars/somecalendar.csv',
                                                     calendar_name='Some Calendar Name')
            cal.id
            >>> 5c1d4904211c0a061bc93013
            cal.name
            >>> Some Calendar Name

            # Creating a calendar without specifying a name
            cal = dr.CalendarFile.create('/home/calendars/somecalendar.csv')
            cal.id
            >>> 5c1d4904211c0a061bc93012
            cal.name
            >>> somecalendar.csv

            # Creating a calendar with multiseries id columns
            cal = dr.CalendarFile.create('/home/calendars/somemultiseriescalendar.csv',
                                         calendar_name='Some Multiseries Calendar Name',
                                         multiseries_id_columns=['series_id'])
            cal.id
            >>> 5da9bb21962d746f97e4daee
            cal.name
            >>> Some Multiseries Calendar Name
            cal.multiseries_id_columns
            >>> ['series_id']
        """

        # make sure it's a valid filename, and set the calendar name if not provided
        if isinstance(file_path, str) and os.path.isfile(file_path):
            if not calendar_name:
                calendar_name = os.path.basename(file_path)
        else:
            raise ValueError(f"The provided file does not exist: {file_path}")
        cls._validate_calendar_name(calendar_name)

        form_data = None
        if multiseries_id_columns:
            cls._validate_multiseries_id_columns(multiseries_id_columns)
            form_data = {"multiseries_id_columns": (None, json.dumps(multiseries_id_columns))}

        upload_response = cls._client.build_request_with_file(
            method="post",
            url=cls._upload_url,
            fname=calendar_name,
            file_path=file_path,
            form_data=form_data,
        )
        new_calendar_url = wait_for_async_resolution(
            cls._client, upload_response.headers["Location"]
        )

        return cls.from_location(new_calendar_url)

    @classmethod
    def create_calendar_from_dataset(
        cls,
        dataset_id: str,
        dataset_version_id: Optional[str] = None,
        calendar_name: Optional[str] = None,
        multiseries_id_columns: Optional[List[str]] = None,
        delete_on_error: Optional[bool] = False,
    ) -> CalendarFile:
        """
        Creates a calendar using the given dataset. For information about calendar files, see the
        :ref:`calendar documentation <calendar-files>`

        The provided dataset have the following format:

        .. code-block:: text

            Date,   Event,          Series ID,    Event Duration
            <date>, <event_type>,   <series id>,  <event duration>
            <date>, <event_type>,              ,  <event duration>

        The "Series ID" and "Event Duration" columns are optional.

        Once the CalendarFile has been created, pass its ID with
        the :class:`DatetimePartitioningSpecification <datarobot.DatetimePartitioningSpecification>`
        when setting the target for a time series project in order to use it.

        Parameters
        ----------
        dataset_id : string
            The identifier of the dataset from which to create the calendar.
        dataset_version_id : string, optional
            The identifier of the dataset version from which to create the calendar.
        calendar_name : string, optional
            A name to assign to the calendar. Defaults to the name of the dataset if not provided.
        multiseries_id_columns : list of Optional[str]
            A list of the names of multiseries id columns to define which series an event
            belongs to. Currently only one multiseries id column is supported.
        delete_on_error : boolean, optional
            Whether delete calendar file from Catalog if it's not valid.

        Returns
        -------
        calendar_file : CalendarFile
            Instance with initialized data.

        Raises
        ------
        AsyncProcessUnsuccessfulError
            Raised if there was an error processing the provided calendar file.

        Examples
        --------
        .. code-block:: python

            # Creating a calendar from a dataset
            dataset = dr.Dataset.create_from_file('/home/calendars/somecalendar.csv')
            cal = dr.CalendarFile.create_calendar_from_dataset(
                dataset.id, calendar_name='Some Calendar Name'
            )
            cal.id
            >>> 5c1d4904211c0a061bc93013
            cal.name
            >>> Some Calendar Name

            # Creating a calendar from a new dataset version
            new_dataset_version = dr.Dataset.create_version_from_file(
                dataset.id, '/home/calendars/anothercalendar.csv'
            )
            cal = dr.CalendarFile.create(
                new_dataset_version.id, dataset_version_id=new_dataset_version.version_id
            )
            cal.id
            >>> 5c1d4904211c0a061bc93012
            cal.name
            >>> anothercalendar.csv
        """
        if calendar_name is not None:
            cls._validate_calendar_name(calendar_name)

        payload: Dict[str, Union[Optional[str], bool, List[str]]] = {
            "name": calendar_name,
            "dataset_id": dataset_id,
            "dataset_version_id": dataset_version_id,
        }

        if delete_on_error:
            payload["delete_on_error"] = delete_on_error

        if multiseries_id_columns:
            cls._validate_multiseries_id_columns(multiseries_id_columns)
            payload["multiseries_id_columns"] = multiseries_id_columns

        generation_response = cls._client.post(
            cls._from_dataset_url, data=payload, headers={"Content-Type": "application/json"}
        )
        generated_calendar_url = wait_for_async_resolution(
            cls._client, generation_response.headers["Location"]
        )

        return cls.from_location(generated_calendar_url)

    @classmethod
    def create_calendar_from_country_code(
        cls, country_code: str, start_date: datetime, end_date: datetime
    ) -> CalendarFile:
        """
        Generates a calendar based on the provided country code and dataset start date and end
        dates. The provided country code should be uppercase and 2-3 characters long. See
        :meth:`CalendarFile.get_allowed_country_codes
        <datarobot.CalendarFile.get_allowed_country_codes>` for a list of allowed country codes.

        Parameters
        ----------
        country_code : string
            The country code for the country to use for generating the calendar.
        start_date : datetime.datetime
            The earliest date to include in the generated calendar.
        end_date : datetime.datetime
            The latest date to include in the generated calendar.

        Returns
        -------
        calendar_file : CalendarFile
            Instance with initialized data.
        """
        payload = {
            "countryCode": country_code,
            "startDate": start_date,
            "endDate": end_date,
        }
        generation_response = cls._client.post(
            cls._from_country_code_url, data=payload, headers={"Content-Type": "application/json"}
        )
        generated_calendar_url = wait_for_async_resolution(
            cls._client, generation_response.headers["Location"]
        )
        return cls.from_location(generated_calendar_url)

    @classmethod
    def get_allowed_country_codes(
        cls, offset: Optional[int] = None, limit: Optional[int] = None
    ) -> List[CountryCode]:
        """
        Retrieves the list of allowed country codes that can be used for generating the preloaded
        calendars.

        Parameters
        ----------
        offset : int
            Optional, defaults to 0. This many results will be skipped.
        limit : int
            Optional, defaults to 100, maximum 1000. At most this many results are returned.

        Returns
        -------
        list
            A list dicts, each of which represents an allowed country codes. Each item has the
            following structure:

            * ``name`` : (str) The name of the country.
            * ``code`` : (str) The code for this country. This is the value that should be supplied
              to :meth:`CalendarFile.create_calendar_from_country_code
              <datarobot.CalendarFile.create_calendar_from_country_code>`.
        """
        params = {}
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        return list(unpaginate(cls._allowed_countries_list_url, params, cls._client))

    @classmethod
    def get(cls, calendar_id: str) -> CalendarFile:
        """
        Gets the details of a calendar, given the id.

        Parameters
        ----------
        calendar_id : str
            The identifier of the calendar.

        Returns
        -------
        calendar_file : CalendarFile
            The requested calendar.

        Raises
        ------
        DataError
            Raised if the calendar_id is invalid, i.e. the specified CalendarFile does not exist.

        Examples
        --------
        .. code-block:: python

            cal = dr.CalendarFile.get(some_calendar_id)
            cal.id
            >>> some_calendar_id
        """
        return cls.from_location(cls._calendar_url.format(calendar_id))

    @classmethod
    def list(
        cls, project_id: Optional[str] = None, batch_size: Optional[int] = None
    ) -> List[CalendarFile]:
        """
        Gets the details of all calendars this user has view access for.

        Parameters
        ----------
        project_id : Optional[str]
            If provided, will filter for calendars associated only with the specified project.
        batch_size : Optional[int]
            The number of calendars to retrieve in a single API call. If specified, the client may
            make multiple calls to retrieve the full list of calendars. If not specified, an
            appropriate default will be chosen by the server.

        Returns
        -------
        calendar_list : list of :class:`CalendarFile <datarobot.CalendarFile>`
            A list of CalendarFile objects.

        Examples
        --------
        .. code-block:: python

            calendars = dr.CalendarFile.list()
            len(calendars)
            >>> 10
        """
        if project_id is not None:
            list_url = cls._base_url + f"?projectId={project_id}"
        else:
            list_url = cls._base_url

        params = {}
        if batch_size is not None:
            params["limit"] = batch_size

        return [cls.from_server_data(entry) for entry in unpaginate(list_url, params, cls._client)]

    @classmethod
    def delete(cls, calendar_id: str) -> None:
        """
        Deletes the calendar specified by calendar_id.

        Parameters
        ----------
        calendar_id : str
            The id of the calendar to delete.
            The requester must have OWNER access for this calendar.

        Raises
        ------
        ClientError
            Raised if an invalid calendar_id is provided.

        Examples
        --------
        .. code-block:: python

            # Deleting with a valid calendar_id
            status_code = dr.CalendarFile.delete(some_calendar_id)
            status_code
            >>> 204
            dr.CalendarFile.get(some_calendar_id)
            >>> ClientError: Item not found
        """
        cls._client.delete(cls._calendar_url.format(calendar_id))

    @classmethod
    def update_name(cls, calendar_id: str, new_calendar_name: str) -> int:
        """
        Changes the name of the specified calendar to the specified name.
        The requester must have at least READ_WRITE permissions on the calendar.

        Parameters
        ----------
        calendar_id : str
            The id of the calendar to update.
        new_calendar_name : str
            The new name to set for the specified calendar.

        Returns
        -------
        status_code : int
            200 for success

        Raises
        ------
        ClientError
            Raised if an invalid calendar_id is provided.

        Examples
        --------
        .. code-block:: python

            response = dr.CalendarFile.update_name(some_calendar_id, some_new_name)
            response
            >>> 200
            cal = dr.CalendarFile.get(some_calendar_id)
            cal.name
            >>> some_new_name

        """
        cls._validate_calendar_name(new_calendar_name)

        response_data: Response = cls._client.patch(
            cls._calendar_url.format(calendar_id), data={"name": new_calendar_name}
        )
        return response_data.status_code

    @classmethod
    def share(cls, calendar_id: str, access_list: List[SharingAccess]) -> int:
        """
        Shares the calendar with the specified users, assigning the specified roles.

        Parameters
        ----------
        calendar_id : str
            The id of the calendar to update
        access_list:
            A list of dr.SharingAccess objects. Specify `None` for the role to delete a user's
            access from the specified CalendarFile. For more information on specific access levels,
            see the :ref:`sharing <sharing>` documentation.

        Returns
        -------
        status_code : int
            200 for success

        Raises
        ------
        ClientError
            Raised if unable to update permissions for a user.
        AssertionError
            Raised if access_list is invalid.

        Examples
        --------
        .. code-block:: python

            # assuming some_user is a valid user, share this calendar with some_user
            sharing_list = [dr.SharingAccess(some_user_username,
                                             dr.enums.SHARING_ROLE.READ_WRITE)]
            response = dr.CalendarFile.share(some_calendar_id, sharing_list)
            response.status_code
            >>> 200

            # delete some_user from this calendar, assuming they have access of some kind already
            delete_sharing_list = [dr.SharingAccess(some_user_username,
                                                    None)]
            response = dr.CalendarFile.share(some_calendar_id, delete_sharing_list)
            response.status_code
            >>> 200

            # Attempt to add an invalid user to a calendar
            invalid_sharing_list = [dr.SharingAccess(invalid_username,
                                                     dr.enums.SHARING_ROLE.READ_WRITE)]
            dr.CalendarFile.share(some_calendar_id, invalid_sharing_list)
            >>> ClientError: Unable to update access for this calendar

        """
        # ensure access_list is a list
        assert isinstance(access_list, list), "access_list must be a list"
        # ensure each item in access_list is a SharingAccess object
        assert all(
            isinstance(access, SharingAccess) for access in access_list
        ), "access_list must be a list of dr.SharingAccess objects"

        payload = {"users": [access.collect_payload() for access in access_list]}
        response_data: Response = cls._client.patch(
            cls._access_control_url.format(calendar_id), data=payload, keep_attrs={"role"}
        )
        return response_data.status_code

    @classmethod
    def get_access_list(
        cls, calendar_id: str, batch_size: Optional[int] = None
    ) -> List[SharingAccess]:
        """
        Retrieve a list of users that have access to this calendar.

        Parameters
        ----------
        calendar_id : str
            The id of the calendar to retrieve the access list for.
        batch_size : Optional[int]
            The number of access records to retrieve in a single API call. If specified, the client
            may make multiple calls to retrieve the full list of calendars. If not specified, an
            appropriate default will be chosen by the server.

        Returns
        -------
        access_control_list : list of :class:`SharingAccess <datarobot.SharingAccess>`
            A list of :class:`SharingAccess <datarobot.SharingAccess>` objects.

        Raises
        ------
        ClientError
            Raised if user does not have access to calendar or calendar does not exist.
        """

        params = {}
        if batch_size is not None:
            params["limit"] = batch_size
        url = cls._access_control_url.format(calendar_id)
        return [
            SharingAccess.from_server_data(datum) for datum in unpaginate(url, params, cls._client)
        ]

    @classmethod
    def _validate_calendar_name(cls, calendar_name: str) -> None:
        if not isinstance(calendar_name, str):
            raise ValueError("Invalid calendar name, only strings are supported")
        try:
            calendar_name.encode("ascii")
        # Which exception we get here depends on whether the input was string or unicode
        # (we allow both).
        except (UnicodeEncodeError, UnicodeDecodeError):
            raise errors.IllegalFileName

    @classmethod
    def _validate_multiseries_id_columns(cls, multiseries_id_columns: Iterable[str]) -> None:
        if not isinstance(multiseries_id_columns, (list, tuple)):
            raise ValueError(
                "Expected list of str for multiseries_id_columns, got: {}".format(
                    multiseries_id_columns
                )
            )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import base64
from concurrent.futures import as_completed, ThreadPoolExecutor
import multiprocessing as mp
from os import path
from typing import Any, Dict, List, Optional, Sequence, Tuple

import requests
from requests.exceptions import RequestException

from datarobot.enums import DEFAULT_MAX_WAIT, DEFAULT_TIMEOUT, FileLocationType
from datarobot.errors import ContentRetrievalTerminatedError
from datarobot.helpers.image_utils import format_image_bytes, ImageOptions


def get_bytes_for_path(
    location: str, continue_on_error: bool = False, base_path: Optional[str] = None, **kwargs: Any
) -> Optional[bytes]:
    """Return file content for path as bytes"""
    try:
        location = path.join(base_path, location) if base_path else location
        with open(location, mode="rb") as f:
            buffer = f.read()
    except OSError:
        buffer = None
        # exception encountered during processing of single row should stop processing
        if not continue_on_error:
            msg = (
                f"Process terminated. Could not retrieve resource: `{location}`. "
                f"When using relative path, location is evaluated against the "
                f"directory where the script is running. Please use `base_path` "
                f"parameter to specify different base directory."
            )
            raise ContentRetrievalTerminatedError(msg)
    return buffer


def get_bytes_for_url(  # type: ignore[return]
    location: str,
    continue_on_error: bool = False,
    headers: Optional[Dict[str, str]] = None,
    **kwargs: Any,
) -> bytes:
    """Return file content for url as bytes"""
    try:
        response = requests.get(url=location, headers=headers, timeout=DEFAULT_TIMEOUT.READ)
        response.raise_for_status()
        return response.content
    except RequestException:
        # exception encountered during processing of single row should stop processing
        if not continue_on_error:
            msg = f"Process terminated. Could not retrieve resource: {location}"
            raise ContentRetrievalTerminatedError(msg)


get_bytes_switcher = {
    FileLocationType.PATH: get_bytes_for_path,
    FileLocationType.URL: get_bytes_for_url,
}


def _get_base64_for_location(
    location: str,
    location_type: str,
    continue_on_error: bool,
    image_options: Optional[ImageOptions] = None,
    **kwargs: Any,
) -> Optional[str]:
    """Retrieve content from the specified location and return it as a base64 string"""
    result = None
    get_bytes_method = get_bytes_switcher[location_type]
    content_bytes = get_bytes_method(location, continue_on_error, **kwargs)  # type: ignore[operator]
    if content_bytes:
        if image_options:
            content_bytes = format_image_bytes(content_bytes, image_options)
        result = base64.b64encode(content_bytes).decode("utf-8")
    # if no content raise exception unless user chosen otherwise
    if not result and not continue_on_error:
        msg = f"Process terminated. Could not retrieve resource: {location}"
        raise ContentRetrievalTerminatedError(msg)
    return result


def _get_base64_for_location_indexed(
    idx: int, *args: Any, **kwargs: Any
) -> Tuple[int, Optional[str]]:
    """Return indexed result of content from specified location converted to base64 string"""
    return idx, _get_base64_for_location(*args, **kwargs)


def _process_locations(
    locations: Sequence[str], n_threads: Optional[int] = None, **kwargs: Any
) -> List[Optional[str]]:
    """Process and return base64 encoded strings for all locations preserving order"""

    # If not specified determine number of threads to use
    if not n_threads:
        try:
            n_threads = mp.cpu_count()
        except Exception:
            n_threads = 4

    if n_threads == 1:
        return [_get_base64_for_location(location, **kwargs) for location in locations]
    else:
        # NOTE: Since results can be returned by threads in any order depending on how long
        # it takes to load and process specific file. To ensure we will preserve order when
        # results are being returned we initialize results array and place results using
        # their indexes.
        results: List[Optional[str]] = [None] * len(locations)

        with ThreadPoolExecutor(max_workers=n_threads) as executor:
            futures = []
            for idx, loc in enumerate(locations):
                futures.append(
                    executor.submit(_get_base64_for_location_indexed, idx, loc, **kwargs)
                )

            for future in as_completed(futures, timeout=DEFAULT_MAX_WAIT):
                try:
                    result_idx, result_value = future.result()
                    results[result_idx] = result_value
                except Exception as exc:
                    # clean up and reraise thread exception
                    executor.shutdown(wait=False)
                    for victim in futures:
                        victim.cancel()
                    raise exc

    return results


def get_encoded_image_contents_from_urls(
    urls: Sequence[str],
    custom_headers: Optional[Dict[str, str]] = None,
    image_options: Optional[ImageOptions] = None,
    continue_on_error: bool = False,
    n_threads: Optional[int] = None,
) -> List[Optional[str]]:
    """
    Returns base64 encoded string of images located in addresses passed in input collection.
    Input collection should hold data of valid image url addresses reachable from
    location where code is being executed. Method will retrieve image, apply specified
    reformatting before converting contents to base64 string. Results will in same
    order as specified in input collection.

    Parameters
    ----------
    urls: Iterable
        Iterable with url addresses to download images from
    custom_headers: dict
        Dictionary containing custom headers to use when downloading files using a URL. Detailed
        data related to supported Headers in HTTP  can be found in the RFC specification for
        headers: https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html
        When used, specified passed values will overwrite default header values.
    image_options: ImageOptions class
        Class holding parameters for use in image transformation and formatting.
    continue_on_error: bool
        If one of rows encounters error while retrieving content (i.e. file does not exist) should
        this error terminate process of downloading consecutive files or should process continue
        skipping this file.
    n_threads: int or None
        Number of threads to use for processing. If "None" is passed, the number of threads is
        determined automatically based on the number of available CPU cores. If this is not
        possible, 4 threads are used.

    Raises
    ------
    ContentRetrievalTerminatedError:
        The error is raised when the flag `continue_on_error` is set to` False` and processing has
        been terminated due to an exception while loading the contents of the file.

    Returns
    -------
    List of base64 encoded strings representing reformatted images.
    """
    return _process_locations(
        locations=urls,
        location_type=FileLocationType.URL,
        image_options=image_options or ImageOptions(),
        headers=custom_headers or {},
        continue_on_error=continue_on_error,
        n_threads=n_threads,
    )


def get_encoded_image_contents_from_paths(
    paths: Sequence[str],
    image_options: Optional[ImageOptions] = None,
    continue_on_error: bool = False,
    n_threads: Optional[int] = None,
    base_path: Optional[str] = None,
) -> List[Optional[str]]:
    """
    Returns base64 encoded string of images located in paths passed in input collection.
    Input collection should hold data of valid image paths reachable from location
    where code is being executed. Method will retrieve image, apply specified
    reformatting before converting contents to base64 string. Results will in same
    order as specified in input collection.

    Parameters
    ----------
    paths: Iterable
        Iterable with path locations to open images from
    image_options: ImageOptions class
        Class holding parameters for image transformation and formatting
    continue_on_error: bool
        If one of rows encounters error while retrieving content (i.e. file does not exist) should
        this error terminate process of downloading consecutive files or should process continue
        skipping this file.
    n_threads: int or None
        Number of threads to use for processing. If "None" is passed, the number of threads is
        determined automatically based on the number of available CPU cores. If this is not
        possible, 4 threads are used.
    base_path: Optional[str]
        Base path to use when opening files. If specified, this path will be used as a base
        directory against which all relative paths will be evaluated. If not specified, the path
        will be evaluated against the directory where this script is running.

    Raises
    ------
    ContentRetrievalTerminatedError:
        The error is raised when the flag `continue_on_error` is set to` False` and processing has
        been terminated due to an exception while loading the contents of the file.

    Returns
    -------
    List of base64 encoded strings representing reformatted images.
    """
    return _process_locations(
        locations=paths,
        location_type=FileLocationType.PATH,
        image_options=image_options or ImageOptions(),
        continue_on_error=continue_on_error,
        n_threads=n_threads,
        base_path=base_path,
    )


def get_encoded_file_contents_from_paths(
    paths: Sequence[str],
    continue_on_error: bool = False,
    n_threads: Optional[int] = None,
    base_path: Optional[str] = None,
) -> List[Optional[str]]:
    """
    Returns base64 encoded string for files located under paths passed in input collection.
    Input collection should hold data of valid file paths locations reachable from
    location where code is being executed. Method will retrieve file and convert its contents
    to base64 string. Results will be returned in same order as specified in input collection.

    Parameters
    ----------
    paths: Iterable
        Iterable with path locations to open images from
    continue_on_error: bool
        If one of rows encounters error while retrieving content (i.e. file does not exist) should
        this error terminate process of downloading consecutive files or should process continue
        skipping this file.
    n_threads: int or None
        Number of threads to use for processing. If "None" is passed, the number of threads is
        determined automatically based on the number of available CPU cores. If this is not
        possible, 4 threads are used.
    base_path: Optional[str]
        Base path to use when opening files. If specified, this path will be used as a base
        directory against which all relative paths will be evaluated. If not specified, the path
        will be evaluated against the directory where this script is running.

    Raises
    ------
    ContentRetrievalTerminatedError:
        The error is raised when the flag `continue_on_error` is set to` False` and processing has
        been terminated due to an exception while loading the contents of the file.

    Returns
    -------
    List of base64 encoded strings representing files.
    """
    return _process_locations(
        locations=paths,
        location_type=FileLocationType.PATH,
        continue_on_error=continue_on_error,
        n_threads=n_threads,
        base_path=base_path,
    )


def get_encoded_file_contents_from_urls(
    urls: Sequence[str],
    custom_headers: Optional[Dict[str, str]] = None,
    continue_on_error: bool = False,
    n_threads: Optional[int] = None,
) -> List[Optional[str]]:
    """
    Returns base64-encoded string for files located in the URL addresses passed on input. Input
    collection holds data of valid file URL addresses reachable from location where code is being
    executed. Method will retrieve file and convert its contents to base64 string. Results will
    be returned in same order as specified in input collection.

    Parameters
    ----------
    urls: Iterable
        Iterable containing URL addresses to download images from.
    custom_headers: dict
        Dictionary with headers to use when downloading files using a URL. Detailed data
        related to supported Headers in HTTP  can be found in the RFC specification:
        https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html. When specified,
        passed values will overwrite default header values.
    continue_on_error: bool
        If a row encounters an error while retrieving content (i.e., file does not exist),
        specifies whether the error results in terminating the process of downloading
        consecutive files or the process continues. Skipped files will be marked as missing.
    n_threads: int or None
        Number of threads to use for processing. If "None" is passed, the number of threads is
        determined automatically based on the number of available CPU cores. If this is not
        possible, 4 threads are used.

    Raises
    ------
    ContentRetrievalTerminatedError:
        The error is raised when the flag `continue_on_error` is set to` False` and processing has
        been terminated due to an exception while loading the contents of the file.

    Returns
    -------
    List of base64 encoded strings representing files.
    """
    return _process_locations(
        locations=urls,
        location_type=FileLocationType.URL,
        headers=custom_headers or {},
        continue_on_error=continue_on_error,
        n_threads=n_threads,
    )



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import os
import shutil
import tempfile
import time

import trafaret as t

from datarobot import errors
from datarobot._compat import Int, String
from datarobot.enums import DEFAULT_MAX_WAIT, EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS
from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate


class ExecutionEnvironmentVersion(APIObject):
    """A version of a DataRobot execution environment.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        the id of the execution environment version
    environment_id: str
        the id of the execution environment the version belongs to
    build_status: str
        the status of the execution environment version build
    label: Optional[str]
        the label of the execution environment version
    description: Optional[str]
        the description of the execution environment version
    created_at: Optional[str]
        ISO-8601 formatted timestamp of when the execution environment version was created
    docker_context_size: Optional[int]
        The size of the uploaded Docker context in bytes if available or None if not
    docker_image_size: Optional[int]
        The size of the built Docker image in bytes if available or None if not
    """

    _path = "executionEnvironments/{}/versions/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("environment_id"): String(),
            t.Key("build_status"): String(),
            t.Key("label", optional=True): t.Or(String(max_length=50, allow_blank=True), t.Null()),
            t.Key("description", optional=True): t.Or(
                String(max_length=10000, allow_blank=True), t.Null()
            ),
            t.Key("created", optional=True) >> "created_at": String(),
            t.Key("docker_context_size", optional=True): t.Or(Int(), t.Null()),
            t.Key("docker_image_size", optional=True): t.Or(Int(), t.Null()),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs):
        self._set_values(**kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.label or self.id!r})"

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        id,
        environment_id,
        build_status,
        label=None,
        description=None,
        created_at=None,
        docker_context_size=None,
        docker_image_size=None,
    ):
        self.id = id
        self.environment_id = environment_id
        self.build_status = build_status
        self.label = label
        self.description = description
        self.created_at = created_at
        self.docker_context_size = docker_context_size
        self.docker_image_size = docker_image_size

    @classmethod
    def create(
        cls,
        execution_environment_id,
        docker_context_path,
        label=None,
        description=None,
        max_wait=DEFAULT_MAX_WAIT,
    ):
        """Create an execution environment version.

        .. versionadded:: v2.21

        Parameters
        ----------
        execution_environment_id: str
            the id of the execution environment
        docker_context_path: str
            the path to a docker context archive or folder
        label: Optional[str]
            short human readable string to label the version
        description: Optional[str]
            execution environment version description
        max_wait: Optional[int]
            max time to wait for a final build status ("success" or "failed").
            If set to None - method will return without waiting.

        Returns
        -------
        ExecutionEnvironmentVersion
            created execution environment version

        Raises
        ------
        datarobot.errors.AsyncTimeoutError
            if version did not reach final state during timeout seconds
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        if os.path.isdir(docker_context_path):
            docker_context_file = tempfile.NamedTemporaryFile(suffix=".zip")
            shutil.make_archive(
                os.path.splitext(docker_context_file.name)[0], "zip", docker_context_path
            )
        else:
            docker_context_file = open(  # pylint: disable=consider-using-with
                docker_context_path, "rb"
            )

        try:
            url = cls._path.format(execution_environment_id)

            response = cls._client.build_request_with_file(
                form_data={"label": label, "description": description},
                fname="docker_context",
                file_field_name="docker_context",
                filelike=docker_context_file,
                url=url,
                method="post",
            )

            version_id = response.json()["id"]

            if max_wait is None:
                return cls.get(execution_environment_id, version_id)
            return cls._await_final_build_status(execution_environment_id, version_id, max_wait)
        finally:
            docker_context_file.close()

    @classmethod
    def list(cls, execution_environment_id, build_status=None):
        """List execution environment versions available to the user.

        .. versionadded:: v2.21

        Parameters
        ----------
        execution_environment_id: str
            the id of the execution environment
        build_status: Optional[str]
            build status of the execution environment version to filter by.
            See datarobot.enums.EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS for valid options

        Returns
        -------
        List[ExecutionEnvironmentVersion]
            a list of execution environment versions.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = cls._path.format(execution_environment_id)
        data = unpaginate(url, {"build_status": build_status}, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, execution_environment_id, version_id):
        """Get execution environment version by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        execution_environment_id: str
            the id of the execution environment
        version_id: str
            the id of the execution environment version to retrieve

        Returns
        -------
        ExecutionEnvironmentVersion
            retrieved execution environment version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = cls._path.format(execution_environment_id)
        path = f"{url}{version_id}/"
        return cls.from_location(path)

    def download(self, file_path):
        """Download execution environment version.

        .. versionadded:: v2.21

        Parameters
        ----------
        file_path: str
            path to create a file with execution environment version content

        Returns
        -------
        ExecutionEnvironmentVersion
            retrieved execution environment version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = self._path.format(self.environment_id)
        path = f"{url}{self.id}/download/"

        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    def get_build_log(self):
        """Get execution environment version build log and error.

        .. versionadded:: v2.21

        Returns
        -------
        Tuple[str, str]
            retrieved execution environment version build log and error.
            If there is no build error - None is returned.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        url = self._path.format(self.environment_id)
        path = f"{url}{self.id}/buildLog/"
        result = self._client.get(path).json()
        log = result["log"]
        error = result["error"]
        if error == "":
            error = None
        return log, error

    def refresh(self):
        """Update execution environment version with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        base_url = self._path.format(self.environment_id)
        url = f"{base_url}{self.id}/"
        response = self._client.get(url)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))

    @classmethod
    def _await_final_build_status(cls, execution_environment_id, version_id, max_wait):
        """Awaits until an execution environment version gets to a final state.

        Parameters
        ----------
        execution_environment_id: str
            the id of the execution environment
        version_id: str
            the id of the execution environment version to retrieve
        max_wait: int or Optional[float]
            max time to wait in seconds

        Returns
        -------
        ExecutionEnvironmentVersion
            execution environment version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        datarobot.errors.AsyncTimeoutError
            if version did not reach final state during timeout seconds
        """
        timeout_at = time.time() + max_wait
        while True:
            version = cls.get(execution_environment_id, version_id)
            if version.build_status in EXECUTION_ENVIRONMENT_VERSION_BUILD_STATUS.FINAL_STATUSES:
                break
            if time.time() >= timeout_at:
                raise errors.AsyncTimeoutError(
                    "Timeout while waiting for environment version to be built. Timeout: {}, "
                    "current state: {}, environment id: {}, environment version id: {}".format(
                        max_wait,
                        version.build_status,
                        version.id,
                        version.environment_id,
                    )
                )
            time.sleep(5)
        return version



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, Iterable, List, Optional, TYPE_CHECKING, Union

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

if TYPE_CHECKING:
    from datarobot.models.api_object import ServerDataType


class ChangeRequest(APIObject):  # pylint: disable=missing-class-docstring
    _path = "changeRequests/"

    _user_operations = t.Dict(
        {
            t.Key("can_update"): t.Bool(),
            t.Key("can_resolve"): t.Bool(),
            t.Key("can_cancel"): t.Bool(),
            t.Key("can_comment"): t.Bool(),
            t.Key("can_review"): t.Bool(),
        }
    ).ignore_extra("*")

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("entity_type"): String(),
            t.Key("entity_id"): String(),
            t.Key("action"): String(),
            t.Key("change"): t.Or(t.Dict(allow_extra="*"), t.Null()),
            t.Key("change_version_id"): String(),
            t.Key("status"): String(),
            t.Key("auto_apply"): t.Bool(),
            t.Key("user_id"): String(),
            t.Key("user_name"): t.Or(String(allow_blank=True), t.Null()),
            t.Key("comment"): String(allow_blank=True),
            t.Key("num_approvals_required"): Int(),
            t.Key("created_at"): String(),
            t.Key("updated_at"): String(),
            t.Key("user_operations"): _user_operations,
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        entity_type: str,
        entity_id: str,
        action: str,
        change: Dict[str, str],
        change_version_id: str,
        status: str,
        auto_apply: bool,
        user_id: str,
        user_name: Optional[str],
        comment: str,
        num_approvals_required: int,
        created_at: str,
        updated_at: str,
        user_operations: Dict[str, bool],
    ) -> None:
        self.id = id
        self.entity_type = entity_type
        self.entity_id = entity_id
        self.action = action
        self.change = change
        self.change_version_id = change_version_id
        self.status = status
        self.auto_apply = auto_apply
        self.user_id = user_id
        self.user_name = user_name
        self.comment = comment
        self.num_approvals_required = num_approvals_required
        self.created_at = created_at
        self.updated_at = updated_at
        self.user_operations = user_operations

    def __repr__(self) -> str:
        return (
            "{}(id={id}, "
            "entity_type={entity_type}, "
            "entity_id={entity_id}, "
            "action={action}, "
            "status={status})"
        ).format(
            self.__class__.__name__,
            id=self.id,
            entity_type=self.entity_type,
            entity_id=self.entity_id,
            action=self.action,
            status=self.status,
        )

    @classmethod
    def from_server_data(
        cls,
        data: ServerDataType,
        keep_attrs: Optional[Iterable[str]] = None,
    ) -> ChangeRequest:
        keep_attrs = set(keep_attrs or [])
        keep_attrs.add("change")
        return super().from_server_data(data, keep_attrs)

    @classmethod
    def list(
        cls,
        entity_type: str,
        entity_id: Optional[str] = None,
        statuses: Optional[List[str]] = None,
    ) -> List[ChangeRequest]:
        """List accessible change requests with the given filters.

        Parameters
        ----------
        entity_type : str
            type of the entity to search change requests for.
        entity_id : Optional[str]
            ID if the entity to filter change requests by.
        statuses : list[str], optional
            list of change request statuses to filter change requests by.

        Returns
        -------
        list[ChangeRequest]
        """
        param: Dict[str, Union[str, List[str]]] = {"entity_type": entity_type}
        if entity_id:
            param["entity_id"] = entity_id
        if statuses:
            param["status"] = statuses

        data = unpaginate(cls._path, param, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, change_request_id: str) -> ChangeRequest:
        """Retrieve change request by ID.

        Parameters
        ----------
        change_request_id : str
            ID of the change request to retrieve.

        Returns
        -------
        ChangeRequest
        """
        path = f"{cls._path}{change_request_id}/"
        return cls.from_location(path)

    @classmethod
    def create(
        cls,
        entity_type: str,
        entity_id: str,
        action: str,
        change: Dict[str, str],
        auto_apply: Optional[bool] = None,
        comment: Optional[str] = None,
    ) -> ChangeRequest:
        """Create new change request.

        Parameters
        ----------
        entity_type : str
            type of the entity to request changes for.
        entity_id : str
            ID of the entity to request changes for.
        action : str
            action the request is intended to perform on the entity.
        change : dict
            change that the user wants to apply to the entity.
        auto_apply : Optional[bool]
            whether the change request should be automatically applied after being approved.
        comment : Optional[str]
            free form text description on the request.

        Returns
        -------
        ChangeRequest
            created change request.
        """
        payload: Dict[str, Union[str, bool, Dict[str, str]]] = {
            "entity_type": entity_type,
            "entity_id": entity_id,
            "action": action,
            "change": change,
        }

        if auto_apply is not None:
            payload["auto_apply"] = auto_apply

        if comment is not None:
            payload["comment"] = comment

        response = cls._client.post(cls._path, data=payload)
        return cls.from_server_data(response.json())

    def update(
        self,
        change: Optional[Dict[str, str]] = None,
        auto_apply: Optional[bool] = None,
        comment: Optional[str] = None,
    ) -> ChangeRequest:
        """Update the change request.

        Parameters
        ----------
        change : dict, optional
            new change to set for the request.
            Setting a new change updates the change version.
        auto_apply : Optional[bool]
            whether the change request should be automatically applied after being approved.
        comment : Optional[str]
            comment to replace the existing change request comment with.

        Returns
        -------
        ChangeRequest
            an instance of updated change request.
        """
        payload = {
            "change": change,
            "auto_apply": auto_apply,
            "comment": comment,
        }
        payload = {k: v for k, v in payload.items() if v is not None}

        if not payload:
            raise ValueError()

        path = f"{self._path}{self.id}/"
        response = self._client.patch(path, data=payload)
        return self.from_server_data(response.json())

    def _set_status(self, status: str) -> ChangeRequest:
        """Update change request status.

        Returns
        -------
        ChangeRequest
            change request after the status change.
        """
        path = f"{self._path}{self.id}/status/"
        response = self._client.patch(path, data={"status": status})
        wait_for_async_resolution(self._client, response.headers["Location"])
        return self.get(self.id)

    def cancel(self) -> ChangeRequest:
        """Cancel the change request.

        Returns
        -------
        ChangeRequest
            cancelled change request.
        """
        return self._set_status("cancelled")

    def resolve(self) -> ChangeRequest:
        """Resolve the change request.

        Apply the change, requested to the entity.
        Only approved change requests can be resolved.

        Returns
        -------
        ChangeRequest
            resolved change request.
        """
        return self._set_status("resolving")

    def approve(self, comment: Optional[str] = None) -> ChangeRequestReview:
        """Approve the change request.

        A shortcut for creating a review on the change request with `approved` status.

        Parameters
        ----------
        comment : Optional[str]
            free form comment to explain the approval.

        Returns
        -------
        ChangeRequestReview
        """
        return ChangeRequestReview.create(self.id, "approved", self.change_version_id, comment)

    def request_changes(self, comment: Optional[str] = None) -> ChangeRequestReview:
        """Request updates to the change request.

        A shortcut for creating a review on the change request with `changesRequested` status.

        Parameters
        ----------
        comment : Optional[str]
            free form comment to explain why the updates are requested.

        Returns
        -------
        ChangeRequestReview
        """
        return ChangeRequestReview.create(
            self.id, "changesRequested", self.change_version_id, comment
        )


class ChangeRequestReview(APIObject):  # pylint: disable=missing-class-docstring
    _path = "changeRequests/{}/reviews/"
    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("change_request_id"): String(),
            t.Key("change_version_id"): String(),
            t.Key("status"): String(),
            t.Key("comment"): String(allow_blank=True),
            t.Key("user_id"): String(),
            t.Key("user_name"): t.Or(String(allow_blank=True), t.Null()),
            t.Key("created_at"): String(),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        change_request_id: str,
        change_version_id: str,
        status: str,
        comment: str,
        user_id: str,
        user_name: Optional[str],
        created_at: str,
    ) -> None:
        self.id = id
        self.change_request_id = change_request_id
        self.change_version_id = change_version_id
        self.status = status
        self.comment = comment
        self.user_id = user_id
        self.user_name = user_name
        self.created_at = created_at

    def __repr__(self) -> str:
        return (
            "{}(change_request_id={change_request_id}, " "status={status}, " "comment='{comment}')"
        ).format(
            self.__class__.__name__,
            change_request_id=self.change_request_id,
            status=self.status,
            comment=self.comment,
        )

    @classmethod
    def list(cls, change_request_id: str) -> List[ChangeRequestReview]:
        """List change request reviews.

        Parameters
        ----------
        change_request_id : str
            ID of the change request to list reviews for.

        Returns
        -------
        list[ChangeRequestReview]
        """
        param: Dict[Any, Any] = {}
        url = cls._path.format(change_request_id)
        data = unpaginate(url, param, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def create(
        cls,
        change_request_id: str,
        status: str,
        change_version_id: str,
        comment: Optional[str] = None,
    ) -> ChangeRequestReview:
        """Add a review for the change request.

        Parameters
        ----------
        change_request_id : str
            ID of the change request to add review for.
        status : str
            review status to create.
        change_version_id : str
            ID of the changes version the review is associated with.
        comment : Optional[str]
            free form text description on the review.

        Returns
        -------
        ChangeRequestReview
        """
        payload = {
            "change_version_id": change_version_id,
            "status": status,
        }
        if comment is not None:
            payload["comment"] = comment

        url = cls._path.format(change_request_id)
        response = cls._client.post(url, data=payload)
        return cls.from_server_data(response.json())

    @classmethod
    def get(cls, change_request_id: str, review_id: str) -> ChangeRequestReview:
        """Retrieve a review by ID for the given change request.

        Parameters
        ----------
        change_request_id : str
            ID of the change request to retrieve review for.
        review_id : str
            ID of the review to retrieve.

        Returns
        -------
        ChangeRequestReview
        """
        url = cls._path.format(change_request_id)
        path = f"{url}{review_id}/"
        return cls.from_location(path)
